#!/usr/local/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test _Root module."""

from __future__ import absolute_import
import sys
import unittest
from mock import patch, MagicMock as Mock

from DIRAC import S_OK, S_ERROR
from Tests.Utilities.GeneralUtils import assertDiracFailsWith, assertDiracSucceeds

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Interfaces.API.NewInterface.Applications._Root'

# pylint: disable=protected-access


class RootTestCase(unittest.TestCase):
  """Base class for the _Root test cases."""

  def setUp(self):
    """set up the objects."""
    # Mock out modules that spawn other threads
    mocked_modules = {'DIRAC.DataManagementSystem.Client.DataManager': Mock()}
    self.module_patcher = patch.dict(sys.modules, mocked_modules)
    self.module_patcher.start()
    from ILCDIRAC.Interfaces.API.NewInterface.Applications import _Root
    self.root = _Root({})

  def tearDown(self):
    self.module_patcher.stop()

  def test_userjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.root._userjobmodules(module_mock), self)

  def test_prodjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.root._prodjobmodules(module_mock), self)

  def test_userjobmodules_fails(self):
    with patch('%s._setUserJobFinalization' % MODULE_NAME, new=Mock(return_value=S_OK('something'))),\
        patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_test_err'))):
      assertDiracFailsWith(self.root._userjobmodules(None),
                            'userjobmodules failed', self)

  def test_prodjobmodules_fails(self):
    with patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_OK('something'))), \
        patch('%s._setOutputComputeDataList' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_other_test_err'))):
      assertDiracFailsWith(self.root._prodjobmodules(None),
                            'prodjobmodules failed', self)

  def test_setmacro(self):
    assertDiracFailsWith(self.root.setMacro(None), 'not allowed here', self)

  def test_resolvelinkedstepparams(self):
    instance_mock = Mock()
    step_mock = Mock()
    step_mock.getType.return_value = 'abc'
    self.root._inputappstep = None
    self.root._jobsteps = ['', '', step_mock]
    self.root._linkedidx = 2
    assertDiracSucceeds(self.root._resolveLinkedStepParameters(instance_mock), self)
    instance_mock.setLink.assert_called_once_with('InputFile', 'abc', 'OutputFile')

  def test_resolvelinkedstepparams_nothing_happens(self):
    instance_mock = Mock()
    self.root._inputappstep = None
    self.root._jobsteps = None
    self.root._linkedidx = ['abc']
    assertDiracSucceeds(self.root._resolveLinkedStepParameters(instance_mock), self)
    self.assertFalse(instance_mock.setLink.called)

  def test_checkconsistency_noscript(self):
    self.root.script = None
    assertDiracFailsWith(self.root._checkConsistency(), 'script or macro not defined', self)

  def test_checkconsistency_noversion(self):
    self.root.script = 1
    self.root.version = None
    assertDiracFailsWith(self.root._checkConsistency(), 'you need to specify the root version', self)



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""
StdhepCutJava: apply generator level cuts after pythia or whizard
"""
from __future__ import absolute_import
__RCSID__ = "$Id$"

from ILCDIRAC.Interfaces.API.NewInterface.Applications import StdhepCut


class StdhepCutJava(StdhepCut):
  """Call stdhep cut after whizard of pythia.

  Usage:

  >>> py = Pythia()
  ...
  >>> cut = StdhepCutJava()
  >>> cut.getInputFromApp(py)
  >>> cut.setSteeringFile("mycut.cfg")
  >>> cut.setMaxNbEvts(10)
  >>> cut.setNbEvtsPerFile(10)
  """

  def __init__(self, paramdict=None):
    super(StdhepCutJava, self).__init__(paramdict)

    self.appname = 'stdhepcutjava'
    self._modulename = 'StdHepCutJava'
    self._moduledescription = 'Module to cut on Generator (Whizard of PYTHIA) written in java'
    self.datatype = 'gen'



#!/usr/bin/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test the FindSteeringFileDir class."""

from __future__ import absolute_import
import unittest
from mock import patch, MagicMock as Mock

from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Core.Utilities.FindSteeringFileDir import getSteeringFileDir, getSteeringFileDirName
from Tests.Utilities.GeneralUtils import assertDiracFailsWith, assertDiracSucceedsWith_equals

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Core.Utilities.FindSteeringFileDir'


class TestFindSteeringFileDir(unittest.TestCase):
  """Test the different methods of the class."""

  def setUp(self):
    pass

  def test_getsteerfiledirname(self):
    getval_mock = Mock()
    getval_mock.getValue.return_value = 'my_test_getval'
    ops_mock = Mock()
    ops_mock.return_value = getval_mock
    cvmfs_mock = Mock(return_value=S_ERROR('some_err'))
    soft_mock = Mock(return_value=S_OK('my_test_dir_retval'))
    check_mock = Mock(return_value=S_OK('ignoreme'))
    with patch("%s.Operations" % MODULE_NAME, new=ops_mock), \
         patch('%s.checkCVMFS' % MODULE_NAME, new=cvmfs_mock), \
         patch('%s.getSoftwareFolder' % MODULE_NAME, new=soft_mock), \
         patch('%s.check' % MODULE_NAME, new=check_mock):
      assertDiracSucceedsWith_equals(getSteeringFileDirName('mytest_plat', 'myappTestme', 'vTest1.0'),
                                      'my_test_dir_retval', self)
      getval_mock.getValue.assert_called_once_with(
          '/AvailableTarBalls/mytest_plat/myappTestme/vTest1.0/Dependencies/steeringfiles/version', '')

  def test_getsteerfiledirname_fails(self):
    getval_mock = Mock()
    getval_mock.getValue.return_value = ''
    ops_mock = Mock()
    ops_mock.return_value = getval_mock
    with patch("%s.Operations" % MODULE_NAME, new=ops_mock):
      assertDiracFailsWith(getSteeringFileDirName('mytest_plat', 'myappTestme', 'vTest1.0'),
                            'could not find attached steeringfile version', self)

  def test_getsteerfiledir_cvmfs_success(self):
    cvmfs_mock = Mock(return_value=S_OK(['mytestListEntry#1', 'other_entry_dontusethis', '', '18319']))
    with patch('%s.checkCVMFS' % MODULE_NAME, new=cvmfs_mock):
      assertDiracSucceedsWith_equals(getSteeringFileDir('myTestPlatform_1', 'v123Test'),
                                      'mytestListEntry#1', self)
      cvmfs_mock.assert_called_once_with('myTestPlatform_1', ['steeringfiles', 'v123Test'])

  def test_getsteerfiledir_getsoftware_fails(self):
    cvmfs_mock = Mock(return_value=S_ERROR('some_err'))
    soft_mock = Mock(return_value=S_ERROR('software_fails_test'))
    with patch('%s.checkCVMFS' % MODULE_NAME, new=cvmfs_mock), \
         patch('%s.getSoftwareFolder' % MODULE_NAME, new=soft_mock):
      assertDiracFailsWith(getSteeringFileDir('myTestPlatform_1', 'v123Test'),
                            'software_fails_test', self)
      cvmfs_mock.assert_called_once_with('myTestPlatform_1', ['steeringfiles', 'v123Test'])
      soft_mock.assert_called_once_with('myTestPlatform_1', 'steeringfiles', 'v123Test')

  def test_getsteerfiledir_check_fails(self):
    cvmfs_mock = Mock(return_value=S_ERROR('some_err'))
    soft_mock = Mock(return_value=S_OK('softDir_test'))
    check_mock = Mock(return_value=S_ERROR('check_fails_testme'))
    with patch('%s.checkCVMFS' % MODULE_NAME, new=cvmfs_mock), \
         patch('%s.getSoftwareFolder' % MODULE_NAME, new=soft_mock), \
         patch('%s.check' % MODULE_NAME, new=check_mock):
      assertDiracFailsWith(getSteeringFileDir('myTestPlatform_1', 'v123Test'),
                            'check_fails_testme', self)
      cvmfs_mock.assert_called_once_with('myTestPlatform_1', ['steeringfiles', 'v123Test'])
      soft_mock.assert_called_once_with('myTestPlatform_1', 'steeringfiles', 'v123Test')
      check_mock.assert_called_once_with('steeringfiles.v123Test', '.', ['softDir_test'])



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Download the production log files from the Log storage element.

See the JDL of production jobs for the log file location. One can also download a full directory of log files.

Example::

  dirac-ilc-get-prod-log -F /ilc/prod/clic/..../1225_23.tar.gz

Options:
   -D, --LogFileDir lfnDirectory      Production log dir to download
   -F, --LogFile lfn                  Production log to download
   -O, --OutputDir localDir           Output directory (default ./)
   -P, --ProdID prodID                Download the log folder 000 for this production ID
   -A, --All                          Get logs from all sub-directories
   --Query                            Set the metadata query to use for finding production files: Key1:Val1[,K2:V2,...]
   -N, --NoPrompt                     Do not query before download.

:since: Mar 21, 2013
:author: Stephane Poss
"""
from __future__ import absolute_import
__RCSID__ = "$Id$"

import os

from DIRAC.Core.Base import Script
from DIRAC.Core.Utilities.DIRACScript import DIRACScript
from DIRAC import gLogger, S_OK, S_ERROR, exit as dexit
from DIRAC.Core.Utilities.PromptUser import promptUser


class _Params(object):
  """Parameter object."""

  def __init__(self):
    self.logD = ''
    self.logF = ''
    self.outputdir = './'
    self.prodid = ''
    self.query = {}
    self.getAllSubdirs = False
    self.noPromptBeforeDL = False

  def setLogFileD(self, opt):
    self.logD = opt
    return S_OK()

  def setLogFileF(self, opt):
    self.logF = opt
    return S_OK()

  def setOutputDir(self, opt):
    self.outputdir = opt
    return S_OK()

  def setProdID(self, opt):
    self.prodid = opt
    return S_OK()

  def setAllGet(self, _):
    self.getAllSubdirs = True
    return S_OK()

  def setNoPrompt(self, _):
    self.noPromptBeforeDL = True
    return S_OK()

  def setQuery(self, query):
    if ':' not in query:
      return S_ERROR('Query syntax not correct')
    self.query = dict(q.split(':') for q in query.split(','))
    return S_OK()

  def registerSwitches(self):
    """registers switches."""
    Script.registerSwitch('D:', 'LogFileDir=', 'Production log dir to download', self.setLogFileD)
    Script.registerSwitch('F:', 'LogFile=', 'Production log to download', self.setLogFileF)
    Script.registerSwitch('O:', 'OutputDir=', 'Output directory (default %s)' % self.outputdir,
                          self.setOutputDir)
    Script.registerSwitch(
        'P:',
        'ProdID=',
        'Production ID. The folder will be found my metadata query for ProdID',
        self.setProdID)
    Script.registerSwitch('A', 'All', 'Get logs from all sub-directories', self.setAllGet)
    Script.registerSwitch('N', 'NoPrompt', 'No prompt before download', self.setNoPrompt)
    Script.registerSwitch('', 'Query=', 'Overwrite Meta query: Key1:Val1[,Key2:Val2...]', self.setQuery)
    Script.setUsageMessage('%s -F /ilc/prod/.../LOG/.../somefile' % Script.scriptName)


def _printErrorReport(res):
  """Print the failures for the call to getFile or Directory."""
  if res['Value'] and res['Value']['Failed']:
    for lfn in res['Value']['Failed']:
      gLogger.error("%s %s" % (lfn, res['Value']['Failed'][lfn]))
      return S_ERROR()
  return S_OK()


@DIRACScript()
def main():
  """get production log files from LogSE."""
  clip = _Params()
  clip.registerSwitches()
  Script.parseCommandLine()
  if not (clip.logF or clip.logD or clip.prodid):
    Script.showHelp()
    dexit(1)
  from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
  ops = Operations()
  storageElementName = ops.getValue('/LogStorage/LogSE', 'LogSE')
  from DIRAC.Resources.Storage.StorageElement import StorageElementItem as StorageElement
  logSE = StorageElement(storageElementName)

  if clip.prodid and not (clip.logD or clip.logF):
    result = _getLogFolderFromID(clip)
    if not result['OK']:
      gLogger.error(result['Message'])
      dexit(1)

  if clip.logD:
    if not clip.noPromptBeforeDL:
      res = promptUser('Are you sure you want to get ALL the files in this directory?')
      if not res['OK']:
        dexit()
      choice = res['Value']
      if choice.lower() == 'n':
        dexit(0)

    if isinstance(clip.logD, str):
      res = logSE.getDirectory(clip.logD, localPath=clip.outputdir)
      _printErrorReport(res)
    elif isinstance(clip.logD, list):
      for logdir in clip.logD:
        gLogger.notice('Getting log files from ' + str(logdir))
        res = logSE.getDirectory(logdir, localPath=clip.outputdir)
        _printErrorReport(res)

  if clip.logF:
    res = logSE.getFile(clip.logF, localPath=clip.outputdir)
    _printErrorReport(res)


def _getLogFolderFromID(clip):
  """Obtain the folder of the logfiles from the prodID.

  Fills the clip.logD variable
  """
  from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
  from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

  # Check if transformation exists and get its type
  server = TransformationClient()
  result = server.getTransformation(clip.prodid)
  if not result['OK']:
    return result
  transType = result['Value']['Type']
  query = {'ProdID': clip.prodid}
  if 'Reconstruction' in transType:
    query['Datatype'] = 'DST'
  if clip.query:
    query.update(clip.query)
    gLogger.notice('Using query: %r' % query)

  result = FileCatalogClient().findDirectoriesByMetadata(query, '/')
  if not result['OK']:
    return result
  elif result['Value']:
    # return value is dictionary with (directory ID: directory)
    folders = list(result['Value'].values())
    # remove the base directory of the production, keep only 000, 001 etc.
    folders = [folder for folder in folders if not folder.endswith(str(clip.prodid))]
    if not clip.getAllSubdirs:
      folders = folders[:1]
    clip.logD = []
    # drop the trailing, e.g., 00X
    baseLFN = folders[0].rsplit('/', 1)[0]
    for folder in folders:
      subFolderNumber = folder.rsplit('/', 1)[1]
      logdir = os.path.join(baseLFN, 'LOG', subFolderNumber)
      if logdir not in clip.logD:
        gLogger.notice('Setting logdir to %s' % logdir)
        clip.logD.append(logdir)

  else:
    return S_ERROR("Cannot discover the LogFilePath: No output files yet")

  return S_OK()


if __name__ == '__main__':
  main()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Provides a set of methods to prepare the option files needed by the ILC applications.

:author: Stephane Poss
:since: Jan 29, 2010
"""

from __future__ import print_function
from __future__ import absolute_import
import os


from xml.etree.ElementTree import ElementTree
from xml.etree.ElementTree import Element
from xml.etree.ElementTree import Comment
from xml.etree.ElementTree import tostring

from DIRAC import S_OK, gLogger, S_ERROR, gConfig

from ILCDIRAC.Core.Utilities.ResolveDependencies import resolveDeps
from ILCDIRAC.Core.Utilities.PrepareLibs import removeLibc
from ILCDIRAC.Core.Utilities.OverlayFiles import getOverlayFiles
from ILCDIRAC.Core.Utilities.CombinedSoftwareInstallation import getSoftwareFolder
from ILCDIRAC.Core.Utilities.MarlinXML import setOverlayFilesParameter, setOutputFileParameter
from ILCDIRAC.Workflow.Modules.OverlayInput import allowedBkg
import six

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"


def getNewLDLibs(platform, application, applicationVersion):
  """Prepare the LD_LIBRARY_PATH environment variable: make sure all lib folder are included.

  :param str platform: System config used for the job
  :param str application: name of the application considered
  :param str applicationVersion: version of the application considered
  :return: new LD_LIBRARY_PATH
  """
  log = LOG.getSubLogger("GetLDLibs")
  log.verbose("Getting all lib folders")
  new_ld_lib_path = ""
  deps = resolveDeps(platform, application, applicationVersion)
  for dep in deps:
    res = getSoftwareFolder(platform, dep["app"], dep['version'])
    if not res['OK']:
      continue
    basedepfolder = res['Value']
    if os.path.exists(os.path.join(basedepfolder, "lib")):
      log.verbose("Found lib folder in %s" % (basedepfolder))
      newlibdir = os.path.join(basedepfolder, "lib")
      new_ld_lib_path = newlibdir
      # Remove the libc
      removeLibc(new_ld_lib_path)
    if os.path.exists(os.path.join(basedepfolder, "LDLibs")):
      log.verbose("Found lib folder in %s" % (basedepfolder))
      newlibdir = os.path.join(basedepfolder, "LDLibs")
      new_ld_lib_path = newlibdir
      # Remove the libc
      removeLibc(new_ld_lib_path)
  if "LD_LIBRARY_PATH" in os.environ:
    if new_ld_lib_path:
      new_ld_lib_path = new_ld_lib_path + ":%s" % os.environ["LD_LIBRARY_PATH"]
    else:
      new_ld_lib_path = os.environ["LD_LIBRARY_PATH"]
  return new_ld_lib_path


def getNewPATH(platform, application, applicationVersion):
  """Same as :func:`getNewLDLibs`,but for the PATH.

  :param str platform: System config used for the job
  :param str application: name of the application considered
  :param str applicationVersion: version of the application considered
  :return: new PATH
  """
  log = LOG.getSubLogger("GetPaths")
  log.verbose("Getting all PATH folders")
  new_path = ""
  deps = resolveDeps(platform, application, applicationVersion)
  for dep in deps:
    res = getSoftwareFolder(platform, dep['app'], dep['version'])
    if not res['OK']:
      continue
    depfolder = res['Value']
    if os.path.exists(os.path.join(depfolder, "bin")):
      log.verbose("Found bin folder in %s" % (depfolder))
      newpathdir = os.path.join(depfolder, "bin")
      new_path = newpathdir
  if "PATH" in os.environ:
    if new_path:
      new_path = new_path + ":%s" % os.environ["PATH"]
    else:
      new_path = os.environ["PATH"]
  return new_path


def prepareWhizardFile(input_in, evttype, energy, randomseed, nevts, lumi, output_in):
  """Prepares the whizard.in file to run.

  Using specified parameters in the job definition passed from :mod:`~ILCDIRAC.Workflow.Modules.WhizardAnalysis`

  :param str input_in: input whizard.in to modify
  :param str evttype: process type that will prepend stdhep output name
  :param int randomseed: random seed to use
  :param int nevts: number of events to generate
  :param lumi: luminosity to use
  :type lumi: int or float
  :param str output_in: whizard.in output file name (usually whizard.in)
  :return: S_OK
  """
  foundprocessid = False
  with open(input_in, "r") as inputfile, open(output_in, "w") as outputfile:
    for line in inputfile:
      if line.count("seed"):
        outputfile.write(" seed = %s\n" % randomseed)
      elif line.count("sqrts"):
        outputfile.write(" sqrts = %s\n" % energy)
      elif line.count("n_events") and not lumi:
        outputfile.write(" n_events = %s\n" % nevts)
      elif lumi and line.count("luminosity"):
        outputfile.write(" luminosity = %s\n" % lumi)
      elif line.count("write_events_file") and evttype:
        outputfile.write(" write_events_file = \"%s\" \n" % evttype)
      elif line.count("process_id"):
        outputfile.write(line)
        if line.split("\"")[1]:
          foundprocessid = True
      else:
        outputfile.write(line)

  return S_OK(foundprocessid)


def prepareWhizardFileTemplate(input_in, evttype, parameters, output_in):
  """Prepares the whizard.in file to run.

  Using specified parameters in the job definition passed from :mod:`~ILCDIRAC.Workflow.Modules.WhizardAnalysis`

  :param str input_in: input whizard.in to modify
  :param str evttype: process type that will prepend stdhep output name
  :param dict parameters: dictionary of parameters to set in the whizard.in
  :param str output_in: whizard.in output file name (usually whizard.in)
  :return: S_OK()
  """
  foundprocessid = False

  replaceDict = {}
  replaceDict["SEEDSEED"] = " seed = %s\n" % parameters['SEED']
  replaceDict["ENERGYENERGY"] = " sqrts = %s\n" % parameters['ENERGY']
  replaceDict["RECOILRECOIL"] = " beam_recoil = %s\n" % parameters['RECOIL']
  replaceDict["NBEVTSNBEVTS"] = " n_events = %s\n" % parameters['NBEVTS']
  replaceDict["LUMILUMI"] = " luminosity=%s\n" % parameters['LUMI']
  replaceDict["INITIALSINITIALS"] = " keep_initials = %s\n" % parameters['INITIALS']
  replaceDict["PNAME1PNAME1"] = " particle_name = '%s'\n" % parameters['PNAME1']
  replaceDict["PNAME2PNAME2"] = " particle_name = '%s'\n" % parameters['PNAME2']
  replaceDict["POLAB1POLAB1"] = " polarization = %s\n" % parameters['POLAB1']
  replaceDict["POLAB2POLAB2"] = " polarization = %s\n" % parameters['POLAB2']
  replaceDict["USERB1USERB1"] = " USER_spectrum_on = %s\n" % parameters['USERB1']
  replaceDict["USERB2USERB2"] = " USER_spectrum_on = %s\n" % parameters['USERB2']
  replaceDict["USERSPECTRUMB1"] = " USER_spectrum_mode = %s\n" % parameters['USERSPECTRUM']
  replaceDict["USERSPECTRUMB2"] = " USER_spectrum_mode = -%s\n" % parameters['USERSPECTRUM']
  replaceDict["ISRB1ISRB1"] = " ISR_on = %s\n" % parameters['ISRB1']
  replaceDict["ISRB2ISRB2"] = " ISR_on = %s\n" % parameters['ISRB2']
  replaceDict["EPAB1EPAB1"] = " EPA_on = %s\n" % parameters['EPAB1']
  replaceDict["EPAB2EPAB2"] = " EPA_on = %s\n" % parameters['EPAB2']

  with open(input_in, "r") as inputfile, open(output_in, "w") as outputfile:
    for line in inputfile:
      written = False
      for para, value in replaceDict.items():
        if line.count(para):
          outputfile.write(value)
          written = True
          break  # break from looping dict
      if line.count("write_events_file") and evttype:
        outputfile.write(' write_events_file = "%s" \n' % evttype)
      elif line.count("process_id"):
        outputfile.write(line)
        if line.split("\"")[1]:
          foundprocessid = True
      elif not written:
        outputfile.write(line)

  return S_OK(foundprocessid)


def prepareSteeringFile(inputSteering, outputSteering, detectormodel,
                        stdhepFile, mac, nbOfRuns, startFrom,
                        randomseed, mcrunnumber,
                        processID='', debug=False, outputlcio=None,
                        filemeta=None):
  """Writes out a steering file for Mokka.

  Using specified parameters in the job definition passed from :mod:`~ILCDIRAC.Workflow.Modules.MokkaAnalysis`

  :param str inputSteering: input steering file name
  :param str outputSteering: new steering file that will be used by Mokka
  :param str detectormodel: detector model to use from the DB
  :param str stdhepFile: generator file name to put in the mac file, if needed
  :param str mac: input macro file
  :param int nbOfRuns: number of runs to use
  :param int startFrom: First event to read from the generator file
  :param int randomseed: Seed to use
  :param int mcrunnumber: MC Run number written to lcio file header
  :param str processID: process ID written to lcio file header
  :param bool debug: overwrite default print level. If set to True, don't change printLevel steering parameter
  :param str outputlcio: output slcio file name
  :param dict filemeta: meta data dictionary used to set various metadata parameters Mokka can write to the lcio file header
  :return: S_OK()
  """
  if filemeta is None:
    filemeta = {}

  macname = "mokkamac.mac"
  if len(mac) < 1:
    with open(macname, "w") as macfile:
      if stdhepFile:
        macfile.write("/generator/generator %s\n" % stdhepFile)
      macfile.write("/run/beamOn %s\n" % nbOfRuns)
  else:
    macname = mac

  with open(inputSteering, "r") as inputsteer, open(str(outputSteering), "w") as output:
    for line in inputsteer:
      if line.count(
          "/Mokka/init/initialMacroFile") or line.count("/Mokka/init/BatchMode") or line.count("/Mokka/init/randomSeed"):
        continue
      if outputlcio and line.count("lcioFilename"):
        continue
      if not outputlcio and detectormodel and line.count("/Mokka/init/detectorModel"):
        continue
      output.write(line)
    if detectormodel:
      output.write("#Set detector model to value specified\n")
      output.write("/Mokka/init/detectorModel %s\n" % detectormodel)

    if not debug:
      output.write("#Set debug level to 1\n")
      output.write("/Mokka/init/printLevel 1\n")
    output.write("#Set batch mode to true\n")
    output.write("/Mokka/init/BatchMode true\n")
    output.write("#Set mac file to the one created on the site\n")
    output.write("/Mokka/init/initialMacroFile %s\n" % macname)
    output.write("#Setting random seed\n")
    output.write("/Mokka/init/randomSeed %s\n" % (randomseed))
    output.write("#Setting run number, same as seed\n")
    output.write("/Mokka/init/mcRunNumber %s\n" % (mcrunnumber))
    if outputlcio:
      output.write("#Set outputfile name to job specified\n")
      output.write("/Mokka/init/lcioFilename %s\n" % outputlcio)
    if processID:
      output.write("#Set processID as event parameter\n")
      output.write("/Mokka/init/lcioEventParameter string Process %s\n" % processID)
    elif 'GenProcessID' in filemeta:
      output.write("#Set processID as event parameter\n")
      output.write("/Mokka/init/lcioEventParameter string Process %s\n" % filemeta['GenProcessID'])
    if 'CrossSection' in filemeta:
      output.write("/Mokka/init/lcioEventParameter float CrossSection_fb %s\n" % float(filemeta['CrossSection']))
    if 'Energy' in filemeta:
      output.write("/Mokka/init/lcioEventParameter float Energy %s\n" % float(filemeta['Energy']))
    if 'PolarizationB1' in filemeta:
      polb1 = filemeta['PolarizationB1']
      if not polb1.count('L') and not polb1.count('R'):
        polb1 = '0.'
      else:
        polb1 = polb1.replace("L", "-").replace("R", "")
        if polb1 == '-':
          polb1 = '-1.0'
        elif polb1 == '':
          polb1 = '1.0'
        else:
          polb1 = str(float(polb1) / 100.)
      output.write("/Mokka/init/lcioEventParameter float Pol_ep %s\n" % float(polb1))
    if 'PolarizationB2' in filemeta:
      polb2 = filemeta['PolarizationB2']
      if not polb2.count('L') and not polb2.count('R'):
        polb2 = '0.'
      else:
        polb2 = polb2.replace("L", "-").replace("R", "")
        if polb2 == '-':
          polb2 = '-1.0'
        elif polb2 == '':
          polb2 = '1.0'
        else:
          polb2 = str(float(polb2) / 100.)
      output.write("/Mokka/init/lcioEventParameter float Pol_em %s\n" % float(polb2))

    output.write("#Set event start number to value given as job parameter\n")
    output.write("/Mokka/init/startEventNumber %d\n" % startFrom)

  return S_OK(True)


def fixedXML(element):
  """As the ElementTree writes out proper XML, we need to corrupt it for LCFI."""
  fixed_element = element.decode().replace("&amp;", "&")
  fixed_element = fixed_element.replace("&gt;", ">").replace("&lt;", "<")
  return fixed_element


def prepareXMLFile(finalxml, inputXML, inputGEAR, inputSLCIO,
                   numberofevts, outputFile, outputREC, outputDST, debug,
                   dd4hepGeoFile=None,
                   overlayParam=None,
                  ):
  """Write out a xml file for Marlin.

  Takes in input the specified job parameters for Marlin application given from :mod:`~ILCDIRAC.Workflow.Modules.MarlinAnalysis`

  :param str finalxml: name of the xml file that will be used by Marlin
  :param str inputXML: name of the provided input XML file
  :param str inputGEAR: name of the Gear file
  :param inputSLCIO: input slcio file list
  :type inputSLCIO: list(str)
  :param int numberofevts: number of events to process
  :param str outputFile: name of the outputfile
  :param str outputREC: file name of REC
  :param str outputDST: file name of DST
  :param bool debug: set to True to use given mode, otherwise set verbosity to SILENT
  :param str dd4hepGeoFile: path to the dd4hep Geometry XML file, optional, default None
  :param int overlayParam: list of tuples of background type, number of events in each background file, and processorName; optional, default None
  :return: S_OK
  """
  tree = ElementTree()
  try:
    tree.parse(inputXML)
  except Exception as x:
    LOG.error("Found Exception when parsing Marlin input XML", repr(x))
    return S_ERROR("Found Exception when parsing Marlin input XML")

  # Handle inputSLCIO being list or string
  if isinstance(inputSLCIO, list):
    inputSLCIO = " ".join(inputSLCIO)
  elif not isinstance(inputSLCIO, six.string_types):
    return S_ERROR("inputSLCIO is neither string nor list! Actual type is %s " % type(inputSLCIO))

  glob = tree.find('global')
  lciolistfound = False
  for param in glob.findall("parameter"):  # pylint: disable=E1101
    if param.attrib.get('name') == 'LCIOInputFiles' and inputSLCIO:
      lciolistfound = True
      com = Comment("input file list changed")
      glob.insert(0, com)  # pylint: disable=E1101
      param.text = inputSLCIO
    if numberofevts > 0 and param.attrib.get('name') == 'MaxRecordNumber':
      if 'value' in param.attrib:
        param.attrib['value'] = str(numberofevts)
        com = Comment("MaxRecordNumber changed")
        glob.insert(0, com)  # pylint: disable=E1101
    if param.attrib.get('name') == "GearXMLFile":
      if 'value' in param.attrib:
        param.attrib['value'] = inputGEAR
        com = Comment("input gear changed")
        glob.insert(0, com)  # pylint: disable=E1101
      else:
        param.text = str(inputGEAR)
        com = Comment("input gear changed")
        glob.insert(0, com)  # pylint: disable=E1101
    if not debug:
      if param.attrib.get('name') == 'Verbosity':
        param.text = "SILENT"
        com = Comment("verbosity changed")
        glob.insert(0, com)  # pylint: disable=E1101

  # Add lcioInputFiles parameter if it was not present before
  if not lciolistfound and inputSLCIO:
    name = {"name": "LCIOInputFiles"}
    lciolist = Element("parameter", name)
    lciolist.text = inputSLCIO
    globparams = tree.find("global")
    globparams.append(lciolist)  # pylint: disable=E1101

  resOF = setOutputFileParameter(tree, outputFile, outputREC, outputDST)
  if not resOF['OK']:
    return resOF
  resOver = setOverlayFilesParameter(tree, overlayParam)
  if not resOver['OK']:
    return resOver

  for param in tree.findall('processor'):
    if 'name' not in param.attrib:
      continue
    # Deal with the InitializeDD4hep parameter value for the XML File
    if param.attrib.get('type') == "InitializeDD4hep" and dd4hepGeoFile is not None:
      for subparam in param.findall('parameter'):
        if subparam.attrib.get('name') == "DD4hepXMLFile":
          subparam.text = dd4hepGeoFile
          com = Comment("DD4hepGeoFile changed")
          param.insert(0, com)

  # now, we need to de-escape some characters as otherwise LCFI goes crazy because it does not unescape
  root = tree.getroot()
  root_str = fixedXML(tostring(root))
  with open(finalxml, "w") as of:
    of.write(root_str)
  # tree.write(finalxml)
  return S_OK(True)


def prepareMacFile(inputmac, outputmac, stdhep, nbevts,
                   startfrom, detector=None, randomseed=0,
                   outputlcio=None, _debug=False):
  """Writes out a mac file for SLIC.

  Takes the parameters passed from :mod:`~ILCDIRAC.Workflow.Modules.SLICAnalysis` to define a new mac file if none was provided

  :param str inputmac: name of the specified mac file
  :param str outputmac: name of the final mac file used by SLIC
  :param str stdhep: name of the generator file to use
  :param int nbevts: number of events to process
  :param int startfrom: event nu,ber to start from in the generator file
  :param str detector: Detector model to use.
  :param int randomseed: random seed to use for simulation
  :param str outputlcio: name of the produced output slcio file, this is useful when combined with :func:`setOutputData() <ILCDIRAC.Interfaces.API.NewInterface.UserJob.UserJob.setOutputData>`
  :param bool debug: UNUSED
  :return: S_OK
  """

  listtext = []
  replacelines = []
  replacelines.append("/generator/filename")
  replacelines.append("/generator/skipEvents")
  # replacelines.append("/run/initialize")
  replacelines.append("/random/seed")
  replacelines.append("/lcio/path")
  replacelines.append("/run/beamOn")

  with open(inputmac, 'r') as inputmacfile, open(outputmac, 'w') as output:

    for line in inputmacfile:
      if any(rl in line for rl in replacelines):
        continue
      if detector and line.count("/lcdd/url"):
        continue
      if outputlcio and line.count("/lcio/filename"):
        continue

      listtext.append(line)

    finaltext = "\n".join(listtext)
    finaltext += "\n"
    if detector:
      output.write("/lcdd/url %s.lcdd\n" % detector)
    # output.write("/run/initialize\n")
    if outputlcio:
      output.write("/lcio/filename %s\n" % outputlcio)
    output.write("/lcio/runNumber %s\n" % randomseed)
    output.write(finaltext)
    if stdhep:
      output.write("/generator/filename %s\n" % stdhep)
    output.write("/generator/skipEvents %s\n" % startfrom)
    output.write("/random/seed %s\n" % (randomseed))
    output.write("/run/beamOn %s\n" % nbevts)
  return S_OK(True)


def prepareLCSIMFile(inputlcsim, outputlcsim, numberofevents,
                     trackingstrategy, inputslcio, jars=None,
                     cachedir=None, outputFile=None,
                     outputRECFile=None, outputDSTFile=None,
                     debug=False):
  """Writes out a lcsim file for LCSIM.

  Takes the parameters passed from :mod:`~ILCDIRAC.Workflow.Modules.LCSIMAnalysis`

  :param str inputlcsim: name of the provided lcsim
  :param str outputlcsim: name of the lcsim file on which LCSIM is going to run, defined in :mod:`~ILCDIRAC.Workflow.Modules.LCSIMAnalysis`
  :param int numberofevents: Number of events to process
  :param str trackingstrategy: trackingstrategy file to use, can be empty
  :param inputslcio: list of slcio files on which LCSIM should run
  :type inputslcio: list(str)
  :param jars: list of jar files that should be added in the classpath definition
  :type jars: list(str)
  :param str cachedir: folder that holds the cache directory, instead of Home
  :param str outputFile: File name of the output
  :param str outputDSTFile: filename of the DST file
  :param str outputRECFile: filename of the REC file
  :param bool debug: By default set verbosity to true

  :return: S_OK(string)
  """
  printtext = ''

  tree = ElementTree()
  try:
    tree.parse(inputlcsim)
  except Exception as x:
    LOG.error("Found Exception %s %s" % (Exception, x))
    return S_ERROR("Found Exception %s %s" % (Exception, x))
  if not inputslcio:
    return S_ERROR("Empty input file list")
  baseelem = tree.getroot()
  if baseelem is None:
    return S_ERROR("Invalid lcsim file structure")
  # handle the input slcio file list
  filesinlcsim = tree.find("inputFiles")
  if filesinlcsim is not None:
    filesinlcsim.clear()  # pylint: disable=E1101
  else:
    baseelem = tree.getroot()
    filesinlcsim = Element("inputFiles")
    baseelem.append(filesinlcsim)

  #set = Element("fileSet")
  for slcio in inputslcio:
    newfile = Element('file')
    newfile.text = slcio
    filesinlcsim.append(newfile)
  # filesinlcsim.append(set)

  if jars:
    classpath = tree.find("classpath")
    if classpath is not None:
      classpath.clear()  # pylint: disable=E1101
    else:
      baseelem = tree.getroot()
      classpath = Element("classpath")
      baseelem.append(classpath)
    for jar in jars:
      newjar = Element("jar")
      newjar.text = jar
      classpath.append(newjar)
  # handle number of events
  if numberofevents:
    nbevts = tree.find("control/numberOfEvents")
    if nbevts is not None:
      nbevts.text = str(numberofevents)
    else:
      control = tree.find('control')
      nbevtselm = Element("numberOfEvents")
      nbevtselm.text = str(numberofevents)
      control.append(nbevtselm)  # pylint: disable=E1101
  # handle verbosity
  if debug:
    debugline = tree.find("control/verbose")
    if debugline is not None:
      debugline.text = 'true'
    else:
      control = tree.find('control')
      debugelem = Element('verbose')
      debugelem.text = 'true'
      control.append(debugelem)  # pylint: disable=E1101

  if cachedir:
    cachedirline = tree.find("control/cacheDirectory")
    if cachedirline is not None:
      cachedirline.text = cachedir
    else:
      control = tree.find('control')
      cachedirelem = Element("cacheDirectory")
      cachedirelem.text = cachedir
      control.append(cachedirelem)  # pylint: disable=E1101

  res = gConfig.getOption("/LocalSite/LcsimPrintEveryEvent", 1)
  lcsimPrintEveryEvent = 1 if not res['OK'] else res['Value']
  drivers = tree.findall("drivers/driver")
  eventInterval = tree.find("drivers/driver/eventInterval")
  if eventInterval is not None:
    evtint = eventInterval.text  # pylint: disable=E1101
    if int(evtint) < 10:
      eventInterval.text = "%s" % lcsimPrintEveryEvent
  else:
    notdriver = True
    for driver in drivers:
      if 'type' in driver.attrib:
        if driver.attrib["type"] == "org.lcsim.job.EventMarkerDriver":
          eventInterval = Element("eventInterval")
          eventInterval.text = "%s" % lcsimPrintEveryEvent
          driver.append(eventInterval)
          notdriver = False
    if notdriver:
      drivers = tree.find("drivers")
      propdict = {}
      propdict['name'] = 'evtMarker'
      propdict['type'] = 'org.lcsim.job.EventMarkerDriver'
      eventmarker = Element("driver", propdict)
      eventInterval = Element("eventInterval")
      eventInterval.text = "%s" % lcsimPrintEveryEvent
      eventmarker.append(eventInterval)
      drivers.append(eventmarker)  # pylint: disable=E1101
      execut = tree.find("execute")
      if execut is not None:
        evtmarkattrib = {}
        evtmarkattrib['name'] = "evtMarker"
        evtmark = Element("driver", evtmarkattrib)
        execut.append(evtmark)  # pylint: disable=E1101

  #drivers = tree.findall("drivers/driver")

  if trackingstrategy:
    for driver in drivers:
      if 'type' in driver.attrib:
        if driver.attrib['type'] == 'org.lcsim.recon.tracking.seedtracker.steeringwrappers.SeedTrackerWrapper':
          driver.remove(driver.find('strategyFile'))
          strategy = Element("strategyFile")
          strategy.text = trackingstrategy
          driver.append(strategy)

  mark = tree.find("drivers/driver/marker")
  if mark is not None:
    printtext = mark.text  # pylint: disable=E1101
  else:
    for driver in drivers:
      if 'type' in driver.attrib:
        if driver.attrib["type"] == "org.lcsim.job.EventMarkerDriver":
          marker = Element("marker")
          marker.text = "LCSIM"
          driver.append(marker)
          printtext = marker.text

  # Take care of overlay
  for driver in drivers:
    if 'type' in driver.attrib:
      if driver.attrib['type'] == "org.lcsim.util.OverlayDriver":
        # if driver.attrib['name']=="eventOverlay":
        ov_name = driver.find("overlayName")
        bkg_Type = "gghad"
        if ov_name is not None:
          bkg_Type = ov_name.text.lower()
          res = allowedBkg(bkg_Type)
          if not res['OK']:
            return res
        driver.remove(driver.find('overlayFiles'))
        files = getOverlayFiles(bkg_Type)
        if not files:
          return S_ERROR('Could not find any overlay files')
        overlay = Element('overlayFiles')
        overlay.text = "\n".join(files)
        driver.append(overlay)
  # Take care of the output files
  writerfound = False
  recwriterfound = False
  dstwriterfound = False
  for driver in drivers:
    if 'type' in driver.attrib:
      if driver.attrib['type'] == "org.lcsim.util.loop.LCIODriver":
        if driver.attrib['name'] == "Writer":
          if outputFile:
            driver.remove(driver.find('outputFilePath'))
            outputelem = Element("outputFilePath")
            outputelem.text = outputFile
            driver.append(outputelem)
          writerfound = True
          continue
        if driver.attrib['name'] == "RECWriter" and outputRECFile:
          driver.remove(driver.find('outputFilePath'))
          outputelem = Element("outputFilePath")
          outputelem.text = outputRECFile
          driver.append(outputelem)
          recwriterfound = True
          continue
        if driver.attrib['name'] == "DSTWriter" and outputDSTFile:
          driver.remove(driver.find('outputFilePath'))
          outputelem = Element("outputFilePath")
          outputelem.text = outputDSTFile
          driver.append(outputelem)
          dstwriterfound = True
          continue
  if not writerfound and outputFile:
    drivers = tree.find("drivers")
    propdict = {}
    propdict['name'] = 'Writer'
    propdict['type'] = 'org.lcsim.util.loop.LCIODriver'
    output = Element("driver", propdict)
    outputelem = Element("outputFilePath")
    outputelem.text = outputFile
    output.append(outputelem)
    drivers.append(output)  # pylint: disable=E1101
    execut = tree.find("execute")
    if execut is not None:
      outputattrib = {}
      outputattrib['name'] = "Writer"
      outputmark = Element("driver", outputattrib)
      execut.append(outputmark)  # pylint: disable=E1101
  if not recwriterfound and outputRECFile:
    #drivers = tree.find("drivers")
    #propdict = {}
    #propdict['name'] = 'RECWriter'
    #propdict['type'] = 'org.lcsim.util.loop.LCIODriver'
    #output = Element("driver", propdict)
    #outputelem = Element("outputFilePath")
    #outputelem.text = outputRECFile
    # output.append(outputelem)
    # drivers.append(output)
    #execut = tree.find("execute")
    # if(execut):
    #  outputattrib = {}
    #  outputattrib['name'] = "RECWriter"
    #  outputmark = Element("driver", outputattrib)
    #  execut.append(outputmark)
    pass
  if not dstwriterfound and outputDSTFile:
    #drivers = tree.find("drivers")
    #propdict = {}
    #propdict['name'] = 'DSTWriter'
    #propdict['type'] = 'org.lcsim.util.loop.LCIODriver'
    #output = Element("driver", propdict)
    #outputelem = Element("outputFilePath")
    #outputelem.text = outputDSTFile
    # output.append(outputelem)
    # drivers.append(output)
    #execut = tree.find("execute")
    # if(execut):
    #  outputattrib = {}
    #  outputattrib['name'] = "DSTWriter"
    #  outputmark = Element("driver", outputattrib)
    #  execut.append(outputmark)
    pass
  tree.write(outputlcsim)
  return S_OK(printtext)


def prepareTomatoSalad(inputxml, outputxml, inputSLCIO, outputFile, collection):
  """Prepare the proper steering file for Tomato.

  :param str inputxml: name of the xml steering file
  :param str outputxml: name of the final tomato steering file
  :param str inputSLCIO: inputSLCIO
  :param str outputFile: name of the produced output slcio file, this is
      useful when combined with :func:`setOutputData()
      <ILCDIRAC.Interfaces.API.NewInterface.UserJob.UserJob.setOutputData>`
  :param str collection: collection to be analysed

  :return: S_OK
  """
  if not inputxml:
    with open('default.xml', "w") as inputxmlf:
      inputxmlf.write("""
<?xml version="1.0" encoding="us-ascii"?>
<!-- ?xml-stylesheet type="text/xsl" href="http://ilcsoft.desy.de/marlin/marlin.xsl"? -->
<!-- ?xml-stylesheet type="text/xsl" href="marlin.xsl"? -->

<marlin xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://ilcsoft.desy.de/marlin/marlin.xsd">

   <execute>
      <processor name="MyTomatoProcessor"/>
   </execute>

   <global>
      <parameter name="Verbosity" value="ERROR"/>
   </global>

 <processor name="MyTomatoProcessor" type="TomatoProcessor">
 <!--Automated analysis-->
  <!--Name of the MCParticle collection-->
  <parameter name="MCCollectionName" type="string" lcioInType="MCParticle"> MCParticle </parameter>
  <!--Root OutputFile-->
  <parameter name="OutputFile" type="string" value="tomato.root"/>
  <!--verbosity level of this processor ("DEBUG0-4,MESSAGE0-4,WARNING0-4,ERROR0-4,SILENT")-->
  <!--parameter name="Verbosity" type="string" value=""/-->
</processor>

</marlin>
    """)
    inputxml = 'default.xml'
  tree = ElementTree()
  try:
    tree.parse(inputxml)
  except Exception as x:
    LOG.error("Found Exception %r" % x)
    return S_ERROR("Found Exception %r" % x)
  params = tree.findall('global/parameter')
  glob = tree.find('global')
  lciolistfound = False
  for param in params:
    if 'name' in param.attrib:
      if param.attrib['name'] == 'LCIOInputFiles':
        lciolistfound = True
        com = Comment("input file list changed")
        glob.insert(0, com)  # pylint: disable=E1101
        param.text = inputSLCIO
  if not lciolistfound:
    name = {}
    name["name"] = "LCIOInputFiles"
    lciolist = Element("parameter", name)
    lciolist.text = inputSLCIO
    globparams = tree.find("global")
    globparams.append(lciolist)  # pylint: disable=E1101

  params = tree.findall('processor')
  for param in params:
    if 'type' in param.attrib and param.attrib['type'] == 'TomatoProcessor':
      subparams = param.findall('parameter')
      for subparam in subparams:
        if 'name' in subparam.attrib:
          if outputFile and subparam.attrib['name'] == 'OutputFile':
            com = Comment('Outputfile changed')
            param.insert(0, com)
            subparam.text = outputFile
          if collection and subparam.attrib['name'] == 'MCCollectionName':
            com = Comment('Collections to analyse changed')
            param.insert(0, com)
            subparam.text = collection

  tree.write(outputxml)
  return S_OK()



#!/usr/bin/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test the UserJobFinalization module."""

from __future__ import print_function
from __future__ import absolute_import
import os
import unittest
from mock import patch, call, mock_open, MagicMock as Mock

from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Workflow.Modules.UserJobFinalization import UserJobFinalization
from Tests.Utilities.GeneralUtils import assertEqualsImproved, \
    assertDiracFailsWith, assertDiracSucceeds, assertDiracSucceedsWith, \
    assertDiracSucceedsWith_equals, assertMockCalls

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Workflow.Modules.UserJobFinalization'


@patch('%s.gConfig.getValue' % MODULE_NAME, new=Mock(return_value=[]))
class TestUserJobFinalization(unittest.TestCase):
  """Test the UserJobFinalization module."""

  def setUp(self):
    self.ujf = UserJobFinalization()
    self.ujf.defaultOutputSE = []

  @patch.dict(os.environ, {'JOBID': '123956'}, clear=True)
  def test_application_specific_inputs(self):
    self.ujf.userOutputData = 'something1;something2'
    tmpSE = self.ujf.workflow_commons.get('UserOutputSE', '')
    self.ujf.workflow_commons['UserOutputSE'] = []
    report_mock = Mock(return_value=813412)
    self.ujf.jobReport = report_mock
    result = self.ujf.applicationSpecificInputs()
    assertDiracSucceedsWith_equals(result, 'Parameters resolved', self)
    assertEqualsImproved(self.ujf.enable, True, self)
    assertEqualsImproved(self.ujf.jobID, '123956', self)
    assertEqualsImproved(self.ujf.userOutputData, ['something1',
                                                     'something2'], self)
    assertEqualsImproved(self.ujf.userOutputSE, [], self)
    assertEqualsImproved(self.ujf.userOutputPath, '', self)
    assertEqualsImproved(self.ujf.jobReport, report_mock, self)
    self.ujf.workflow_commons['UserOutputSE'] = tmpSE

  def test_application_specific_inputs_corrupt_state(self):
    self.ujf.enable = 123
    log_mock = Mock()
    self.ujf.log = log_mock
    self.ujf.userOutputData = ['testentryOutputData']
    self.ujf.workflow_commons = {
        'UserOutputSE': ' Some TestOutput Storage;and its brother',
        'UserOutputPath': 'testOutputPathWwo'}
    self.ujf.jobID = 87942
    with patch.dict('os.environ', {}, clear=True):
      result = self.ujf.applicationSpecificInputs()
      assertDiracSucceedsWith_equals(result, 'Parameters resolved', self)
      assertEqualsImproved(self.ujf.jobID, 87942, self)
      assertEqualsImproved(self.ujf.enable, False, self)
      assertEqualsImproved(self.ujf.userOutputData, ['testentryOutputData'],
                            self)
      assertEqualsImproved(self.ujf.userOutputSE, ['Some TestOutput Storage',
                                                     'and its brother'], self)
      assertEqualsImproved(self.ujf.userOutputPath, 'testOutputPathWwo', self)

  @patch.dict(os.environ, {}, clear=True)
  def test_application_specific_inputs_jobid_zero(self):
    self.ujf.jobID = 0
    self.ujf.enable = True
    result = self.ujf.applicationSpecificInputs()
    assertDiracSucceedsWith_equals(result, 'Parameters resolved', self)
    assertEqualsImproved(self.ujf.jobID, 0, self)
    assertEqualsImproved(self.ujf.enable, False, self)

  def test_getcurrentowner(self):
    with patch('%s.getProxyInfo' % MODULE_NAME, new=Mock(return_value=S_OK({'myproxydict': True, 'username': 'readThisOut'}))):
      result = self.ujf.getCurrentOwner()
      assertDiracSucceedsWith(result, 'readThisOut', self)

  def test_getcurrentowner_getproxy_fails(self):
    with patch('%s.getProxyInfo' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_getproxy_fails'))):
      result = self.ujf.getCurrentOwner()
      assertDiracFailsWith(result, 'could not obtain proxy information', self)

  def test_getcurrentowner_username_missing(self):
    with patch('%s.getProxyInfo' % MODULE_NAME, new=Mock(return_value=S_OK({'myproxydict': True}))):
      result = self.ujf.getCurrentOwner()
      assertDiracFailsWith(result, 'could not get username from proxy', self)

  def test_getcurrentvo(self):
    with patch('%s.getProxyInfo' % MODULE_NAME, new=Mock(return_value=S_OK({'myproxydict': True, 'group': 'testilc_blabla'}))):
      result = self.ujf.getCurrentVO()
      assertDiracSucceedsWith(result, 'testilc', self)

  def test_getcurrentvo_getproxy_fails(self):
    with patch('%s.getProxyInfo' % MODULE_NAME, new=Mock(return_value=S_ERROR('test_getproxy_fails'))):
      result = self.ujf.getCurrentVO()
      assertDiracFailsWith(result, 'could not obtain proxy information', self)

  def test_getcurrentvo_group_missing(self):
    with patch('%s.getProxyInfo' % MODULE_NAME, new=Mock(return_value=S_OK({'myproxydict': True}))):
      result = self.ujf.getCurrentVO()
      assertDiracFailsWith(result, 'could not get group from proxy', self)

  def test_constructoutputlfns(self):
    self.ujf.jobID = 0
    self.ujf.workflow_commons = {}
    self.ujf.userOutputData = 'testOutputData'
    self.ujf.userOutputPath = 'testOutputPath'
    with patch('%s.getProxyInfo' % MODULE_NAME, new=Mock(return_value=S_OK({'username': 'myTestUser123', 'group': 'TeILCst_bla'}))), \
         patch('%s.constructUserLFNs' % MODULE_NAME, new=Mock(return_value=S_OK('constructedLFNs'))) as construct_mock:
      result = self.ujf.constructOutputLFNs()
      assertDiracSucceedsWith(result, 'constructedLFNs', self)
      construct_mock.assert_called_once_with(12345, 'TeILCst', 'myTestUser123',
                                              'testOutputData', 'testOutputPath')
      assertEqualsImproved(self.ujf.jobID, 12345, self)

  def test_constructoutputlfns_proxy_missing(self):
    self.ujf.jobID = 8234
    self.ujf.workflow_commons = {}
    with patch('%s.getProxyInfo' % MODULE_NAME, new=Mock(return_value=S_ERROR('proxy_err_testme'))):
      result = self.ujf.constructOutputLFNs()
      assertDiracFailsWith(result, 'could not obtain owner from proxy', self)
      assertEqualsImproved(self.ujf.jobID, 8234, self)

  def test_constructoutputlfns_vo_missing(self):
    self.ujf.workflow_commons = {}
    with patch('%s.getProxyInfo' % MODULE_NAME, new=Mock(return_value=S_OK({'username': 'user943'}))):
      result = self.ujf.constructOutputLFNs()
      assertDiracFailsWith(result, 'could not obtain VO from proxy', self)

  def test_constructoutputlfns_construction_fails(self):
    self.ujf.jobID = 0
    self.ujf.workflow_commons = {}
    self.ujf.userOutputData = 'testOutputData'
    self.ujf.userOutputPath = 'testOutputPath'
    with patch('%s.getProxyInfo' % MODULE_NAME, new=Mock(return_value=S_OK({'username': 'myTestUser123', 'group': 'TeILCst_bla'}))), \
         patch('%s.constructUserLFNs' % MODULE_NAME, new=Mock(return_value=S_ERROR('construction_failed_testerr'))) as construct_mock:
      result = self.ujf.constructOutputLFNs()
      assertDiracFailsWith(result, 'construction_failed_testerr', self)
      construct_mock.assert_called_once_with(12345, 'TeILCst', 'myTestUser123', 'testOutputData', 'testOutputPath')

  def test_transfer_register_and_failover_files(self):
    self.ujf.failoverSEs = ['testSite1', 'testSite2', 'myTestSite', 'privateSite4', 'etc']
    self.ujf.userFileCatalog = 'blaCatalogTestme'
    transfer_mock = Mock()
    transfer_mock.transferAndRegisterFileFailover.return_value = S_OK({})
    filesUploaded = []
    with patch('%s.random.shuffle' % MODULE_NAME) as shuffle_mock:
      result = self.ujf.transferRegisterAndFailoverFiles(transfer_mock, {
          'filename1test': {'metadatadict1': True,
                             'resolvedSE': ('etc', 'privateSite4'),
                             'localpath': '/my/local/path/Testme.txt',
                             'lfn': 'LFN:/ilc/mytest/LFN.txt', 'filedict': 89546},
          'my_other_testfile': {
              'metadatadict1': True, 'resolvedSE': ('testSite1', 'privateSite4'),
              'localpath': '/my/local/path/Testme2.txt', 'lfn': 'LFN:/ilc/othertest/new_lfn.lfn',
              'filedict': 475}}, filesUploaded)
      assertDiracSucceeds(result, self)
      assertEqualsImproved(result['Value'], dict(cleanUp=False), self)
      assertMockCalls(shuffle_mock, [['testSite1', 'testSite2', 'myTestSite', 'privateSite4', 'etc'],
                                       ['testSite1', 'testSite2', 'myTestSite', 'privateSite4', 'etc']],
                       self)
      assertMockCalls(transfer_mock.transferAndRegisterFileFailover,
                      [call('my_other_testfile', '/my/local/path/Testme2.txt',
                            'LFN:/ilc/othertest/new_lfn.lfn', 'testSite1',
                            ['testSite2', 'myTestSite', 'privateSite4', 'etc'],
                            fileCatalog='blaCatalogTestme', fileMetaDict=475),
                       call('filename1test', '/my/local/path/Testme.txt', 'LFN:/ilc/mytest/LFN.txt',
                            'etc', ['testSite1', 'testSite2', 'myTestSite', 'privateSite4'],
                            fileCatalog='blaCatalogTestme', fileMetaDict=89546),
                       ], self)
      self.assertIn('LFN:/ilc/mytest/LFN.txt', filesUploaded)
      self.assertIn('LFN:/ilc/othertest/new_lfn.lfn', filesUploaded)

  def test_transferreg_transfer_fails(self):
    self.ujf.failoverSEs = ['testSite1', 'testSite2', 'myTestSite',
                             'privateSite4', 'etc']
    self.ujf.userFileCatalog = 'blaCatalogTestme'
    transfer_mock = Mock()
    transfer_mock.transferAndRegisterFileFailover.return_value = S_ERROR('some_transfer_test_err')
    filesUploaded = []
    with patch('%s.random.shuffle' % MODULE_NAME) as shuffle_mock:
      result = self.ujf.transferRegisterAndFailoverFiles(transfer_mock, {
          'filename1test': {'metadatadict1': True,
                             'resolvedSE': ('nonexistant', ),
                             'localpath': '/my/local/path/Testme.txt',
                             'lfn': 'LFN:/ilc/mytest/LFN.txt', 'filedict': 89546
                          }}, filesUploaded)
      assertDiracSucceeds(result, self)
      assertEqualsImproved(result['Value'], dict(cleanUp=True), self)
      shuffle_mock.assert_called_once_with(['testSite1', 'testSite2',
                                              'myTestSite', 'privateSite4',
                                              'etc'])
      transfer_mock.transferAndRegisterFileFailover.assert_called_once_with(
          'filename1test', '/my/local/path/Testme.txt', 'LFN:/ilc/mytest/LFN.txt',
          'nonexistant', ['testSite1', 'testSite2', 'myTestSite', 'privateSite4',
                          'etc'], fileCatalog='blaCatalogTestme',
          fileMetaDict=89546)
      assertEqualsImproved(filesUploaded, [], self)

  def test_transferreg_failoverses_empty(self):
    self.ujf.failoverSEs = ['justOneSite']
    self.ujf.userFileCatalog = 'blaCatalogTestme'
    transfer_mock = Mock()
    transfer_mock.transferAndRegisterFileFailover.return_value = S_ERROR('some_transfer_test_err')
    filesUploaded = []
    with patch('%s.random.shuffle' % MODULE_NAME) as shuffle_mock:
      result = self.ujf.transferRegisterAndFailoverFiles(transfer_mock, {
          'filename1test': {'metadatadict1': True,
                             'resolvedSE': ('justOneSite', ),
                             'localpath': '/my/local/path/Testme.txt',
                             'lfn': 'LFN:/ilc/mytest/LFN.txt',
                             'filedict': 89546}}, filesUploaded)
      assertDiracSucceeds(result, self)
      assertEqualsImproved(result['Value'], dict(cleanUp=True), self)
      shuffle_mock.assert_called_once_with(['justOneSite'])
      self.assertFalse(transfer_mock.transferAndRegisterFileFailover.called)
      assertEqualsImproved(filesUploaded, [], self)

  # pylint: disable=unused-argument
  def test_transferreg_some_files_fail(self):
    self.ujf.failoverSEs = ['onlyAvailableSE']
    self.ujf.userFileCatalog = 'myCat'
    transfer_mock = Mock()
    def my_transfer_mock(fileName, localpath, lfn, targetSE, failoverSEs,
                          fileCatalog=None, fileMetaDict=None):
      """Return S_ERROR for the appropriate file from the dictionary and S_OK for the rest.

      Necssary since dictionary iteration order is random
      """
      return TRANSFER_AND_REGISTER_DICT[fileName]
    transfer_mock.transferAndRegisterFileFailover.side_effect = my_transfer_mock
    filesUploaded = []
    with patch('%s.random.shuffle' % MODULE_NAME) as shuffle_mock:
      result = self.ujf.transferRegisterAndFailoverFiles(transfer_mock, {
          'noMoreSEs': {'metadatadict1': True,
                         'resolvedSE': ('onlyAvailableSE',),
                         'localpath': '/doesntmatter',
                         'lfn': 'doesntmattereither', 'filedict': 32},
          'fail_transfer': {
              'metadatadict1': True, 'resolvedSE': ('someOtherSite', ),
              'localpath': '/matterdoesnt', 'lfn': 'neither', 'filedict': 92},
          'workingFile1': {'resolvedSE': ('someOtherSite', ),
                           'localpath': '/my/local/first/path',
                           'lfn': 'LFN:/ilc/some/dir/file1.txt',
                           'filedict': 8520}, 'thisFileWorks.too':
          {'resolvedSE': ('someOtherOtherSite', ),
              'localpath': '/dir/current/local.lfn',
              'lfn': 'LFN:/ilc/mydir/file2.ppt', 'filedict': 98453}
          }, filesUploaded)
      assertDiracSucceeds(result, self)
      assertEqualsImproved(result['Value'], dict(cleanUp=True), self)
      assertMockCalls(shuffle_mock, [['onlyAvailableSE']] * 4, self)
      assertMockCalls((transfer_mock.transferAndRegisterFileFailover),
                      [call('workingFile1', '/my/local/first/path', 'LFN:/ilc/some/dir/file1.txt', 'someOtherSite',
                            ['onlyAvailableSE'], fileCatalog='myCat', fileMetaDict=8520),
                       call('fail_transfer', '/matterdoesnt', 'neither', 'someOtherSite', ['onlyAvailableSE'],
                            fileCatalog='myCat', fileMetaDict=92),
                       call('thisFileWorks.too', '/dir/current/local.lfn', 'LFN:/ilc/mydir/file2.ppt',
                            'someOtherOtherSite', ['onlyAvailableSE'], fileCatalog='myCat', fileMetaDict=98453)],
                      self)
      assertEqualsImproved(filesUploaded, ['LFN:/ilc/some/dir/file1.txt', 'LFN:/ilc/mydir/file2.ppt'], self)


@patch('%s.UserJobFinalization.logWorkingDirectory' % MODULE_NAME, new=Mock())
class TestExecute(unittest.TestCase):
  """Test the execute() method."""

  def setUp(self):
    self.ujf = UserJobFinalization()
    self.ujf.defaultOutputSE = []
    self.ujf.step_commons = {'STEP_NUMBER': '42'}
    self.ujf.workflow_commons = {'TotalSteps': '42'}

  def test_execute(self):
    exists_dict = {'list_of.txt': True, 'filenames.jar': True}
    self.ujf.jobID = 512
    request_mock = Mock()
    request_mock.RequestName = 'job_512_request.xml'
    request_mock.JobID = 512
    request_mock.SourceComponent = "Job_512"
    self.ujf.workflow_commons = {'TotalSteps': '42', 'Request': request_mock,
                                  'JOB_ID': '512', 'IS_PROD': True,
                                  'PRODUCTION_ID': '98245',
                                  'Platform': 'myTestPlatform',
                                  'Owner': 'myTestOwner123RichGuy',
                                  'VO': 'myTestVirtualOrga',
                                  'UserOutputSE': 'myTestReceivingSE'}
    transfer_mock = Mock()
    transfer_mock.transferAndRegisterFile.return_value = S_OK({})
    transfer_mock.transferAndRegisterFileFailover.return_value = S_OK()
    dataman_mock = Mock()
    dataman_mock.replicateAndRegister.return_value = S_OK()
    self.ujf.workflowStatus = S_OK()
    self.ujf.stepStatus = S_OK()
    self.ujf.userOutputData = ['list_of.txt', 'filenames.jar']
    self.ujf.userOutputSE = 'thisValueIsntUsed'
    self.ujf.userOutputPath = 'my/User/OPPath'
    with patch('%s.constructUserLFNs' % MODULE_NAME, new=Mock(return_value=S_OK(['/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/list_of.txt', '/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/filenames.jar']))) as constructlfn_mock, \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])) as exists_mock, \
         patch('%s.open' % MODULE_NAME, mock_open(read_data='myfilecontent,makeniceadlerchecksum')), \
         patch('%s.os.path.getsize' % MODULE_NAME, new=Mock(return_value=3048)), \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/mycurdirTestMe')), \
         patch('%s.getDestinationSEList' % MODULE_NAME, new=Mock(return_value=S_OK(['CERN-DIP-4']))), \
         patch('%s.FailoverTransfer' % MODULE_NAME, new=transfer_mock), \
         patch('%s.DataManager' % MODULE_NAME, new=dataman_mock), \
         patch('%s.time.sleep' % MODULE_NAME, new=Mock(return_value=True)):
      result = self.ujf.execute()
    assertDiracSucceeds(result, self)
    constructlfn_mock.assert_called_once_with(
        512, 'myTestVirtualOrga', 'myTestOwner123RichGuy',
        ['list_of.txt', 'filenames.jar'], 'my/User/OPPath')
    transfer_mock = transfer_mock()  # Necessary for the assumptions
    transfer_mock.transferAndRegisterFile.assert_any_call(
        'list_of.txt', '/mycurdirTestMe/list_of.txt',
        '/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/list_of.txt',
        ['myTestReceivingSE', 'CERN-DIP-4'], fileCatalog=['FileCatalog'],
        fileMetaDict={
            'Status': 'Waiting', 'ADLER32': False, 'ChecksumType': 'ADLER32',
            'Checksum': False,
            'LFN': '/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/list_of.txt',
            'GUID': None, 'Addler': False, 'Size': 3048})
    transfer_mock.transferAndRegisterFile.assert_called_with(
        'filenames.jar', '/mycurdirTestMe/filenames.jar',
        '/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/filenames.jar',
        ['myTestReceivingSE', 'CERN-DIP-4'], fileCatalog=['FileCatalog'],
        fileMetaDict={'Status': 'Waiting', 'ADLER32': False,
                      'ChecksumType': 'ADLER32', 'Checksum': False,
                      'LFN': '/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/filenames.jar',
                      'GUID': None, 'Addler': False, 'Size': 3048})
    assertEqualsImproved(len(transfer_mock.transferAndRegisterFile.mock_calls), 14, self)
    self.assertFalse(transfer_mock.transferAndRegisterFileFailover.called)
    dataman_mock = dataman_mock()
    assertMockCalls(dataman_mock.replicateAndRegister,
                     [('/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/filenames.jar',
                        'myTestReceivingSE'),
                       ('/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/list_of.txt',
                         'myTestReceivingSE')], self, only_these_calls=False)
    assertEqualsImproved(len(exists_mock.mock_calls), 2, self)

  def test_execute_not_last_step(self):
    self.ujf.step_commons = {'STEP_NUMBER': '1'}
    self.ujf.workflow_commons = {'TotalSteps': '10000'}
    result = self.ujf.execute()
    assertDiracSucceeds(result, self)

  def test_execute_resolveIV_fails(self):
    with patch('%s.UserJobFinalization.applicationSpecificInputs' % MODULE_NAME, new=Mock(return_value=S_ERROR('ASI_err_my_test'))):
      result = self.ujf.execute()
      assertDiracFailsWith(result, 'asi_err_my_test', self)

  def test_execute_workflow_status_bad(self):
    self.ujf.stepStatus = S_ERROR('asdjufe')
    result = self.ujf.execute()
    assertDiracSucceedsWith_equals(result, 'No output data upload attempted', self)

  def test_execute_no_data_to_upload(self):
    self.ujf.userOutputData = None
    result = self.ujf.execute()
    assertDiracSucceedsWith_equals(result, 'No output data to upload', self)

  def test_execute_getcandidatefiles_fails(self):
    self.ujf.workflow_commons = {'TotalSteps': '42', 'Owner': 'something',
                                  'VO': 'something_else'}
    self.ujf.userOutputData = ['mylist']
    with patch('%s.UserJobFinalization.getCandidateFiles' % MODULE_NAME, new=Mock(return_value=S_ERROR('get_candidate_files_test_err'))), \
         patch('%s.UserJobFinalization.setApplicationStatus' % MODULE_NAME) as set_app_status_mock:
      result = self.ujf.execute()
      assertDiracSucceeds(result, self)
      set_app_status_mock.assert_called_once_with('get_candidate_files_test_err')

  def test_execute_getdestinationlist_fails(self):
    self.ujf.workflow_commons = {'TotalSteps': '42', 'Owner': 'something',
                                  'VO': 'something_else'}
    self.ujf.userOutputData = ['mylist']
    with patch('%s.UserJobFinalization.setApplicationStatus' % MODULE_NAME) as set_app_status_mock, \
         patch('%s.getDestinationSEList' % MODULE_NAME, new=Mock(return_value=S_ERROR('my_destinationselist_test_err'))), \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(return_value=True)), \
         patch('%s.os.path.getsize' % MODULE_NAME, new=Mock(return_value=13283)):
      result = self.ujf.execute()
      assertDiracFailsWith(result, 'my_destinationselist_test_err', self)
      set_app_status_mock.assert_called_once_with('Failed To Resolve OutputSE')

  def test_execute_constructlfn_fails(self):
    self.ujf.workflow_commons = {'TotalSteps': '42', 'JOB_ID': '512',
                                  'IS_PROD': True, 'PRODUCTION_ID': '98245',
                                  'Platform': 'myTestPlatform',
                                  'Owner': 'myTestOwner123RichGuy',
                                  'VO': 'myTestVirtualOrga'}
    self.ujf.userOutputData = ['list_of.txt', 'filenames.jar']
    with patch('%s.constructUserLFNs' % MODULE_NAME, new=Mock(return_value=S_OK(['/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/list_of.txt', '/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/filenames.jar']))), \
         patch('%s.UserJobFinalization.constructOutputLFNs' % MODULE_NAME, new=Mock(return_value=S_ERROR('my_construct_lfn_err_test'))):
      result = self.ujf.execute()
      assertDiracFailsWith(result, 'my_construct_lfn_err_test', self)

  def test_execute_getfilemetadata_fails(self):
    exists_dict = {'list_of.txt': True, 'something': True}
    self.ujf.jobID = 512
    request_mock = Mock()
    request_mock.RequestName = 'job_512_request.xml'
    request_mock.JobID = 512
    request_mock.SourceComponent = "Job_512"
    report_mock = Mock()
    self.ujf.jobReport = report_mock
    self.ujf.workflow_commons = {'TotalSteps': '42', 'Request': request_mock,
                                  'JOB_ID': '512', 'IS_PROD': True,
                                  'PRODUCTION_ID': '98245', 'Platform': 'myTestPlatform',
                                  'Owner': 'myTestOwner123RichGuy', 'VO': 'myTestVirtualOrga',
                                  'JobReport': report_mock}
    self.ujf.workflowStatus = S_OK()
    self.ujf.stepStatus = S_OK()
    self.ujf.userOutputData = ['something']
    self.ujf.userOutputSE = 'myTestReceivingSE'
    self.ujf.userOutputPath = 'my/User/OPPath'
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])), \
         patch('%s.open' % MODULE_NAME, mock_open(read_data='myfilecontent,makeniceadlerchecksum')), \
         patch('%s.os.path.getsize' % MODULE_NAME, new=Mock(return_value=3048)), \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/mycurdirTestMe')), \
         patch('%s.getDestinationSEList' % MODULE_NAME, new=Mock(return_value=S_OK(['CERN-DIP-4']))), \
         patch('%s.UserJobFinalization.getFileMetadata' % MODULE_NAME, new=Mock(return_value=S_ERROR('my_getfilemeta_test_Err'))):
      result = self.ujf.execute()
      assertDiracSucceedsWith_equals(result, None, self)
      report_mock.setApplicationStatus.assert_called_once_with(
          'my_getfilemeta_test_Err', True)

  def test_execute_getfilemetadata_empty(self):
    exists_dict = {'list_of.txt': True, 'something': True}
    self.ujf.jobID = 512
    request_mock = Mock()
    request_mock.RequestName = 'job_512_request.xml'
    request_mock.JobID = 512
    request_mock.SourceComponent = "Job_512"
    report_mock = Mock()
    self.ujf.jobReport = report_mock
    self.ujf.workflow_commons = {'TotalSteps': '42', 'Request': request_mock,
                                  'JOB_ID': '512', 'IS_PROD': True,
                                  'PRODUCTION_ID': '98245', 'Platform': 'myTestPlatform',
                                  'Owner': 'myTestOwner123RichGuy', 'VO': 'myTestVirtualOrga',
                                  'JobReport': report_mock}
    self.ujf.workflowStatus = S_OK()
    self.ujf.stepStatus = S_OK()
    self.ujf.userOutputData = ['something']
    self.ujf.userOutputSE = 'myTestReceivingSE'
    self.ujf.userOutputPath = 'my/User/OPPath'
    with patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])), \
         patch('%s.open' % MODULE_NAME, mock_open(read_data='myfilecontent,makeniceadlerchecksum')), \
         patch('%s.os.path.getsize' % MODULE_NAME, new=Mock(return_value=3048)), \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/mycurdirTestMe')), \
         patch('%s.getDestinationSEList' % MODULE_NAME, new=Mock(return_value=S_OK(['CERN-DIP-4']))), \
         patch('%s.UserJobFinalization.getFileMetadata' % MODULE_NAME, new=Mock(return_value=S_OK())):
      result = self.ujf.execute()
      assertDiracSucceedsWith_equals(result, None, self)
      report_mock.setApplicationStatus.assert_called_once_with(
          'No Output Data Files To Upload', True)

  def test_execute_ignoreapperrors(self):
    exists_dict = {'list_of.txt': True, 'something': True}
    self.ujf.jobID = 512
    self.ujf.ignoreapperrors = True
    request_mock = Mock()
    request_mock.RequestName = 'job_512_request.xml'
    request_mock.JobID = 512
    request_mock.SourceComponent = "Job_512"
    self.ujf.workflow_commons = {'TotalSteps': '42', 'Request': request_mock,
                                  'JOB_ID': '512', 'IS_PROD': True,
                                  'PRODUCTION_ID': '98245', 'Platform': 'myTestPlatform',
                                  'Owner': 'myTestOwner123RichGuy', 'VO': 'myTestVirtualOrga',
                                  'UserOutputSE': 'myTestReceivingSE'}
    transfer_mock = Mock()
    transfer_mock.transferAndRegisterFile.return_value = S_OK({})
    transfer_mock.transferAndRegisterFileFailover.return_value = S_OK()
    dataman_mock = Mock()
    dataman_mock.replicateAndRegister.return_value = S_OK()
    self.ujf.workflowStatus = S_OK()
    self.ujf.stepStatus = S_OK()
    self.ujf.userOutputData = ['list_of.txt', 'filenames.jar']
    self.ujf.userOutputSE = 'thisValueIsntUsed'
    self.ujf.userOutputPath = 'my/User/OPPath'
    with patch('%s.constructUserLFNs' % MODULE_NAME, new=Mock(return_value=S_OK(['/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/list_of.txt', '/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/filenames.jar']))) as constructlfn_mock, \
         patch('%s.UserJobFinalization.getCandidateFiles' % MODULE_NAME, new=Mock(return_value={'OK': False, 'Value': {}})) as getcf_mock, \
         patch('%s.os.path.exists' % MODULE_NAME, new=Mock(side_effect=lambda path: exists_dict[path])), \
         patch('%s.open' % MODULE_NAME, mock_open(read_data='myfilecontent,makeniceadlerchecksum')), \
         patch('%s.os.path.getsize' % MODULE_NAME, new=Mock(return_value=3048)), \
         patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/mycurdirTestMe')), \
         patch('%s.getDestinationSEList' % MODULE_NAME, new=Mock(return_value=S_OK(['CERN-DIP-4']))), \
         patch('%s.FailoverTransfer' % MODULE_NAME, new=transfer_mock), \
         patch('%s.DataManager' % MODULE_NAME, new=dataman_mock), \
         patch('%s.time.sleep' % MODULE_NAME, new=Mock(return_value=True)), \
         patch('%s.UserJobFinalization.getFileMetadata' % MODULE_NAME, new=Mock(return_value={'OK': False, 'Value': {'workingFile1': {'resolvedSE': ('someOtherSite', ), 'localpath': '/my/local/first/path', 'lfn': 'LFN:/ilc/some/dir/file1.txt', 'filedict': 8520}, 'thisFileWorks.too': {'resolvedSE': ('someOtherOtherSite', ), 'localpath': '/dir/current/local.lfn', 'lfn': 'LFN:/ilc/mydir/file2.ppt', 'filedict': 98453}}})) as getfmd_mock:
      # TODO: does this (getcandidatefiles mock, getfilemetadata mock) make
      # sense? S_ERROR usually will not have a value, so this would actually
      # throw an error in real environment
      result = self.ujf.execute()
    assertDiracSucceeds(result, self)
    getcf_mock.assert_called_once_with([
        {'outputPath': 'TXT', 'outputFile': 'list_of.txt', 'outputDataSE': ['myTestReceivingSE']},
        {'outputPath': 'JAR', 'outputFile': 'filenames.jar', 'outputDataSE': ['myTestReceivingSE']}
        ], ['/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/list_of.txt',
            '/myTestVirtualOrga/testpre/m/myTestOwner123RichGuy/my/User/OPPath/filenames.jar'], '')
    constructlfn_mock.assert_called_once_with(
        512, 'myTestVirtualOrga', 'myTestOwner123RichGuy',
        ['list_of.txt', 'filenames.jar'], 'my/User/OPPath')
    getfmd_mock.assert_called_once_with({})
    transfer_mock = transfer_mock()  # Necessary for the assumptions
    transfer_mock.transferAndRegisterFile.assert_any_call(
        'workingFile1', '/my/local/first/path', 'LFN:/ilc/some/dir/file1.txt',
        ['myTestReceivingSE', 'CERN-DIP-4'], fileCatalog=['FileCatalog'],
        fileMetaDict=8520)
    transfer_mock.transferAndRegisterFile.assert_called_with(
        'thisFileWorks.too', '/dir/current/local.lfn', 'LFN:/ilc/mydir/file2.ppt',
        ['myTestReceivingSE', 'CERN-DIP-4'], fileCatalog=['FileCatalog'],
        fileMetaDict=98453)
    assertEqualsImproved(len(transfer_mock.transferAndRegisterFile.mock_calls),
                          14, self)
    self.assertFalse(transfer_mock.transferAndRegisterFileFailover.called)
    dataman_mock = dataman_mock()
    assertMockCalls(dataman_mock.replicateAndRegister, [('LFN:/ilc/mydir/file2.ppt', 'myTestReceivingSE'),
                                                          ('LFN:/ilc/some/dir/file1.txt', 'myTestReceivingSE')],
                     self, only_these_calls=False)


TRANSFER_AND_REGISTER_DICT = {'fail_transfer': S_ERROR('myrandomerror'),
                               'workingFile1': S_OK(),
                               'thisFileWorks.too': S_OK()}


def main():
  """Executes all tests from this file, useful for profiling, etc."""
  app_spec_suite = unittest.TestLoader().loadTestsFromTestCase(TestUserJobFinalization)
  execute_suite = unittest.TestLoader().loadTestsFromTestCase(TestExecute)
  userjobsuite = unittest.TestSuite([app_spec_suite, execute_suite])
  print(unittest.TextTestRunner(verbosity=2).run(userjobsuite))


if __name__ == '__main__':
  main()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Linear Collider Application.

Allows setting the Steering File dependency, as well as other LC community things

:author: sposs
:since: Nov 1st, 2013
"""

from __future__ import absolute_import
from ILCDIRAC.Interfaces.API.NewInterface.LCUtilityApplication import LCUtilityApplication
from DIRAC.Core.Workflow.Parameter import Parameter

from DIRAC import S_OK

__RCSID__ = "$Id$"


class LCApplication(LCUtilityApplication):
  """LC specific implementation of the applications."""

  def __init__(self, paramdict=None):
    super(LCApplication, self).__init__(paramdict)
    self.steeringFileVersion = ""
    self.forgetAboutInput = False
    self._importLocation = "ILCDIRAC.Workflow.Modules"

  def setSteeringFileVersion(self, version):
    """Define the SteeringFile version to use.

    :param str version:
    """
    self.steeringFileVersion = version

    return S_OK()

  def setForgetAboutInput(self, flag=True):
    """Do not overwrite the input file set in the SteeringFile.

    :param bool flag: True or False
    """

    self.forgetAboutInput = flag

    return S_OK()

  def _getSpecificAppParameters(self, stepdef):
    """Overload of Application._getSpecificAppParameter."""
    stepdef.addParameter(Parameter("ForgetInput", False, "boolean", "", "", False, False,
                                   "Do not overwrite input steering"))
    if self.steeringFileVersion:
      stepdef.addParameter(Parameter("SteeringFileVers", "", "string", "", "", False, False,
                                     "SteeringFile version to use"))
    return S_OK()

  def _setSpecificAppParameters(self, stepinst):
    """Overload of Application._setSpecificAppParameters."""
    stepinst.setValue("ForgetInput", self.forgetAboutInput)

    if self.steeringFileVersion:
      stepinst.setValue("SteeringFileVers", self.steeringFileVersion)

    return S_OK()

  def _doSomethingWithJob(self):
    """Overloads the Application._doSomethingWithJob."""
    if self.steeringFileVersion:
      self._job._addSoftware("steeringfiles", self.steeringFileVersion)
    return S_OK()
# "



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""FileStatusTransformation Agent performs the following actions:

+-----------------------+------------------+---------------------------+--------------+---------------+---------------+
|  Transformation Type  |  Request Status  | Transformation File Status|    Source    |    Target     |    Action     |
+-----------------------+------------------+---------------------------+--------------+---------------+---------------+
| Moving                | Any              | Problematic / Processed   | Available    | Available     | Retry         |
| Replication / Moving  | Any              | Problematic / Processed   | Available    | Not Available | Retry         |
| Moving                | Failed           | Assigned                  | Available    | Available     | Retry         |
| Replication / Moving  | Failed           | Assigned                  | Available    | Not Available | Retry         |
| Replication           | Failed           | Assigned                  | Available    | Available     | Set PROCESSED |
| Replication / Moving  | Any              | Other than Processed      | Not Available| Available     | Set PROCESSED |
| Replication / Moving  | Any              | Any                       | Not Available| Not Available | Set DELETED   |
+-----------------------+------------------+---------------------------+--------------+---------------+---------------+

* If the action is Retry then the request is reset if it exists in RMS, otherwise the file status is set to unused
* Available means the file exists in File Catalog and also exists physically on Storage Elements
* Not Available means the file doesn't exist in File Catalog or one or more replicas are lost on the Storage Elements
"""

from __future__ import absolute_import
import json
from collections import defaultdict

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.PrettyPrint import printTable
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Utilities.Proxy import UserProxy
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
from DIRAC.Resources.Storage.StorageElement import StorageElement
from six.moves import filter

__RCSID__ = "$Id$"

AGENT_NAME = 'ILCTransformation/FileStatusTransformationAgent'

SET_PROCESSED = 'SET_PROCESSED'
SET_DELETED = 'SET_DELETED'
SET_UNUSED = 'SET_UNUSED'

RETRY = 'retry'
RESET_REQUEST = 'RESET_REQUEST'

REPLICATION_TRANS = 'Replication'
MOVING_TRANS = 'Moving'


class FileStatusTransformationAgent(AgentModule):
  """FileStatusTransformationAgent."""

  def __init__(self, *args, **kwargs):
    AgentModule.__init__(self, *args, **kwargs)
    self.name = 'FileStatusTransformationAgent'
    self.enabled = False
    self.shifterProxy = 'DataManager'
    self.transformationTypes = ["Replication"]
    self.transformationStatuses = ["Active"]
    self.transformationFileStatuses = ["Assigned", "Problematic", "Processed", "Unused"]

    self.addressTo = ["andre.philippe.sailer@cern.ch"]
    self.addressFrom = "ilcdirac-admin@cern.ch"
    self.emailSubject = "FileStatusTransformationAgent"

    self.accounting = defaultdict(list)
    self.errors = []

    self.fcClient = FileCatalogClient()
    self.tClient = TransformationClient()
    self.reqClient = ReqClient()
    self.nClient = NotificationClient()

  def checkFileStatusFuncExists(self, status):
    """returns True/False if a function to check transformation files with a given status exists or not."""
    checkFileStatusFuncName = "check_%s_files" % (status.lower())
    if not (hasattr(self, checkFileStatusFuncName) and callable(getattr(self, checkFileStatusFuncName))):
      self.log.warn("Unable to process transformation files with status ", status)
      return False

    return True

  def beginExecution(self):
    """Reload the configurations before every cycle."""
    self.enabled = self.am_getOption('EnableFlag', False)
    self.transformationTypes = self.am_getOption('TransformationTypes', ["Replication"])
    self.transformationStatuses = self.am_getOption('TransformationStatuses', ["Active"])
    self.transformationFileStatuses = self.am_getOption(
        'TransformationFileStatuses', ["Assigned", "Problematic", "Processed", "Unused"])

    self.addressTo = self.am_getOption('MailTo', ["ilcdirac-admin@cern.ch"])
    self.addressFrom = self.am_getOption('MailFrom', "ilcdirac-admin@cern.ch")

    self.transformationFileStatuses = list(filter(self.checkFileStatusFuncExists, self.transformationFileStatuses))
    self.accounting.clear()

    return S_OK()

  def sendNotification(self, transID, transType=None, sourceSEs=None, targetSEs=None):
    """sends email notification about accounting information of a transformation."""
    if not(self.errors or self.accounting):
      return S_OK()

    emailBody = "Transformation ID: %s\n" % transID
    if transType:
      emailBody += "Transformation Type: %s\n" % transType

    if sourceSEs:
      emailBody += "Source SE: %s\n" % (" ".join(str(source) for source in sourceSEs))

    if targetSEs:
      emailBody += "Target SE: %s\n\n" % (" ".join(str(target) for target in targetSEs))

    rows = []
    for action, transFiles in self.accounting.items():
      emailBody += "Total number of files with action %s: %s\n" % (action, len(transFiles))
      for transFile in transFiles:
        rows.append([[transFile['LFN']], [str(transFile['AvailableOnSource'])],
                     [str(transFile['AvailableOnTarget'])], [transFile['Status']], [action]])

    if rows:
      columns = ["LFN", "Source", "Target", "Old Status", "Action"]
      emailBody += printTable(columns, rows, printOut=False, numbering=False, columnSeparator=' | ')

    if self.errors:
      emailBody += "\n\nErrors:"
      emailBody += "\n".join(self.errors)

    self.log.notice(emailBody)
    subject = "%s: %s" % (self.emailSubject, transID)
    for address in self.addressTo:
      res = self.nClient.sendMail(address, subject, emailBody, self.addressFrom, localAttempt=False)
      if not res['OK']:
        self.log.error("Failure to send Email notification to ", address)
        continue

    self.errors = []
    self.accounting.clear()
    return S_OK()

  def logError(self, errStr, varMsg=''):
    self.log.error(errStr, varMsg)
    self.errors.append(errStr + varMsg)

  def execute(self):
    """main execution loop of Agent."""

    res = self.getTransformations()
    if not res['OK']:
      self.log.error('Failure to get transformations', res['Message'])
      return S_ERROR("Failure to get transformations")

    transformations = res['Value']
    if not transformations:
      self.log.notice('No transformations found with Status %s and Type %s ' %
                      (self.transformationStatuses, self.transformationTypes))
      return S_OK()

    self.log.notice('Will treat %d transformations' % len(transformations))
    self.log.notice('Transformations: %s' % ",".join([str(transformation['TransformationID'])
                                                      for transformation in transformations]))

    for trans in transformations:
      transID = trans['TransformationID']
      if 'SourceSE' not in trans or not trans['SourceSE']:
        self.logError("SourceSE not set for transformation, skip processing, transID: ", "%s" % transID)
        self.sendNotification(transID)
        continue

      if 'TargetSE' not in trans or not trans['TargetSE']:
        self.logError("TargetSE not set for transformation, skip processing, transID: ", "%s" % transID)
        self.sendNotification(transID, sourceSEs=trans['SourceSE'])
        continue

      if 'DataTransType' not in trans:
        self.logError("Transformation Type not set for transformation, skip processing, transID: ", "%s" % transID)
        self.sendNotification(transID, sourceSEs=trans['SourceSE'], targetSEs=trans['TargetSE'])
        continue

      res = self.processTransformation(transID, trans['SourceSE'], trans['TargetSE'], trans['DataTransType'], trans)
      if not res['OK']:
        self.log.error('Failure to process transformation with ID:', transID)
        continue

    return S_OK()

  def getTransformations(self, transID=None):
    """returns transformations of a given type and status."""
    res = None
    if transID:
      res = self.tClient.getTransformations(
          condDict={'TransformationID': transID,
                    'Status': self.transformationStatuses,
                    'Type': self.transformationTypes})
    else:
      res = self.tClient.getTransformations(
          condDict={'Status': self.transformationStatuses, 'Type': self.transformationTypes})

    if not res['OK']:
      return res

    result = res['Value']
    for trans in result:
      res = self.tClient.getTransformationParameters(trans['TransformationID'], ['SourceSE', 'TargetSE'])
      if not res['OK']:
        self.log.error('Failure to get SourceSE and TargetSE parameters for Transformation ID:',
                       trans['TransformationID'])
        continue

      trans['SourceSE'] = eval(res['Value']['SourceSE'])
      trans['TargetSE'] = eval(res['Value']['TargetSE'])

      res = self.getDataTransformationType(trans['TransformationID'])
      if not res['OK']:
        self.log.error('Failure to determine Data Transformation Type', "%s: %s"
                       % (trans['TransformationID'], res['Message']))
        continue

      trans['DataTransType'] = res['Value']

    return S_OK(result)

  def getRequestStatus(self, transID, taskIDs):
    """returns request statuses for a given list of task IDs."""
    res = self.tClient.getTransformationTasks(condDict={'TransformationID': transID, 'TaskID': taskIDs})
    if not res['OK']:
      self.log.error('Failure to get Transformation Tasks for Transformation ID:', transID)
      return res

    result = res['Value']
    requestStatus = {}
    for task in result:
      requestStatus[task['TaskID']] = {'RequestStatus': task['ExternalStatus'], 'RequestID': int(task['ExternalID'])}

    return S_OK(requestStatus)

  def getDataTransformationType(self, transID):
    """returns transformation types Replication/Moving/Unknown for a given transformation."""
    res = self.tClient.getTransformationParameters(transID, 'Body')
    if not res['OK']:
      return res

    # if body is empty then we assume that it is a replication transformation
    if not res['Value']:
      return S_OK(REPLICATION_TRANS)

    replication = False
    rmReplica = False
    try:
      body = json.loads(res['Value'])
      for operation in body:
        if 'ReplicateAndRegister' in operation:
          replication = True
        if 'RemoveReplica' in operation:
          rmReplica = True
    except ValueError:
      if 'ReplicateAndRegister' in res['Value']:
        replication = True
        if 'RemoveReplica' in res['Value']:
          rmReplica = True

    if rmReplica and replication:
      return S_OK(MOVING_TRANS)

    if replication:
      return S_OK(REPLICATION_TRANS)

    return S_ERROR("Unknown Transformation Type '%r'" % res['Value'])

  def setFileStatus(self, transID, transFiles, status):
    """sets transformation file status."""

    lfns = [transFile['LFN'] for transFile in transFiles]
    lfnStatuses = {lfn: status for lfn in lfns}

    if lfnStatuses:
      if self.enabled:
        res = self.tClient.setFileStatusForTransformation(transID, newLFNsStatus=lfnStatuses, force=True)
        if not res['OK']:
          self.logError('Failed to set statuses for LFNs ', "%s" % res['Message'])
          return res

      for transFile in transFiles:
        self.accounting[status].append({'LFN': transFile['LFN'],
                                        'Status': transFile['Status'],
                                        'AvailableOnSource': transFile['AvailableOnSource'],
                                        'AvailableOnTarget': transFile['AvailableOnTarget']})
    return S_OK()

  def selectFailedRequests(self, transFile):
    """returns True if transformation file has a failed request otherwise returns False."""
    res = self.getRequestStatus(transFile['TransformationID'], transFile['TaskID'])
    if not res['OK']:
      self.log.error('Failure to get Request Status for Assigned File')
      return False
    result = res['Value']

    if result[transFile['TaskID']]['RequestStatus'] == 'Failed':
      return True

    return False

  def retryStrategyForFiles(self, transID, transFiles):
    """returns retryStrategy Reset Request if a request is found in RMS, otherwise returns set file status to unused."""
    taskIDs = [transFile['TaskID'] for transFile in transFiles]
    res = self.getRequestStatus(transID, taskIDs)
    if not res['OK']:
      return res
    result = res['Value']
    retryStrategy = defaultdict(dict)
    for taskID in taskIDs:
      if taskID is None:
        self.log.error("Task ID is None", "Transformation: %s\n Files: %r " % (transID, transFiles))
        retryStrategy[None]['Strategy'] = SET_UNUSED
        continue
      res = self.reqClient.getRequest(requestID=result[taskID]['RequestID'])
      if not res['OK']:
        self.log.notice('Request %s does not exist setting file status to unused' % result[taskID]['RequestID'])
        retryStrategy[taskID]['Strategy'] = SET_UNUSED
      else:
        retryStrategy[taskID]['Strategy'] = SET_UNUSED  # RESET_REQUEST
        retryStrategy[taskID]['RequestID'] = result[taskID]['RequestID']

    return S_OK(retryStrategy)

  def check_assigned_files(self, actions, transFiles, transType):
    """treatment for transformation files with assigned status."""
    for transFile in transFiles:
      if transFile['AvailableOnSource'] and transFile['AvailableOnTarget']:
        if transType == REPLICATION_TRANS:
          actions[SET_PROCESSED].append(transFile)
        if transType == MOVING_TRANS:
          actions[RETRY].append(transFile)

      elif transFile['AvailableOnSource'] and not transFile['AvailableOnTarget']:
        actions[RETRY].append(transFile)

      elif not transFile['AvailableOnSource'] and transFile['AvailableOnTarget']:
        actions[SET_PROCESSED].append(transFile)

      else:
        # not on src and target
        actions[SET_DELETED].append(transFile)

  def check_unused_files(self, actions, transFiles, transType):
    """treatment for transformation files with unused status."""
    for transFile in transFiles:
      if not transFile['AvailableOnSource'] and transFile['AvailableOnTarget']:
        actions[SET_PROCESSED].append(transFile)

      if not transFile['AvailableOnSource'] and not transFile['AvailableOnTarget']:
        actions[SET_DELETED].append(transFile)

  def check_processed_files(self, actions, transFiles, transType):
    """treatment for transformation files with processed status."""
    for transFile in transFiles:
      if transFile['AvailableOnSource'] and transFile['AvailableOnTarget'] and transType == MOVING_TRANS:
        actions[RETRY].append(transFile)

      if transFile['AvailableOnSource'] and not transFile['AvailableOnTarget']:
        actions[RETRY].append(transFile)

      if not transFile['AvailableOnSource'] and not transFile['AvailableOnTarget']:
        actions[SET_DELETED].append(transFile)

  def check_problematic_files(self, actions, transFiles, transType):
    """treatment for transformation files with problematic status."""
    for transFile in transFiles:
      if transFile['AvailableOnSource'] and transFile['AvailableOnTarget']:
        if transType == REPLICATION_TRANS:
          actions[SET_PROCESSED].append(transFile)
        if transType == MOVING_TRANS:
          actions[RETRY].append(transFile)

      elif transFile['AvailableOnSource'] and not transFile['AvailableOnTarget']:
        actions[RETRY].append(transFile)

      elif not transFile['AvailableOnSource'] and transFile['AvailableOnTarget']:
        actions[SET_PROCESSED].append(transFile)

      else:
        # not available on source and target
        actions[SET_DELETED].append(transFile)

  def retryFiles(self, transID, transFiles):
    """resubmits request or sets file status to unused based on the retry strategy of transformation file."""
    setFilesUnused = []
    setFilesAssigned = []
    res = self.retryStrategyForFiles(transID, transFiles)
    if not res['OK']:
      self.logError('Failure to determine retry strategy (unused / reset request) for files ', "%s" % res['Message'])
      return res

    retryStrategy = res['Value']
    for transFile in transFiles:
      if retryStrategy[transFile['TaskID']]['Strategy'] != RESET_REQUEST:
        setFilesUnused.append(transFile)
        continue

      requestID = retryStrategy[transFile['TaskID']]['RequestID']
      if self.enabled:
        res = self.reqClient.resetFailedRequest(requestID, allR=True)
        if not res['OK']:
          self.logError('Failed to reset request ', 'ReqID: %s Error: %s' % (requestID, res['Message']))
          continue

        if res['Value'] == "Not reset":
          self.logError('Failed to reset request ', 'ReqID: %s is non-recoverable' % requestID)
          continue

        setFilesAssigned.append(transFile)

        res = self.tClient.setTaskStatus(transID, transFile['TaskID'], 'Waiting')
        if not res['OK']:
          self.logError('Failure to set Waiting status for Task ID: ', "%s %s" % (transFile['TaskID'], res['Message']))
          continue

      self.accounting[RESET_REQUEST].append({'LFN': transFile['LFN'],
                                             'Status': transFile['Status'],
                                             'AvailableOnSource': transFile['AvailableOnSource'],
                                             'AvailableOnTarget': transFile['AvailableOnTarget']})

    if setFilesUnused:
      self.setFileStatus(transID, setFilesUnused, 'Unused')

    if setFilesAssigned:
      self.setFileStatus(transID, setFilesAssigned, 'Assigned')

    return S_OK()

  def applyActions(self, transID, actions):
    """sets new file statuses and resets requests."""
    for action, transFiles in actions.items():
      if action == SET_PROCESSED and transFiles:
        self.setFileStatus(transID, transFiles, 'Processed')

      if action == SET_DELETED and transFiles:
        self.setFileStatus(transID, transFiles, 'Deleted')

      if action == RETRY and transFiles:
        # if there is a request in RMS then reset request otherwise set file status unused
        self.retryFiles(transID, transFiles)

  def existsInFC(self, storageElements, lfns):
    """checks if files have replicas registered in File Catalog for all given storageElements."""
    res = {'Successful': {}, 'Failed': {}}
    self.log.info('Looking for replica information for %s files' % len(lfns))
    for index, chunk in enumerate(breakListIntoChunks(lfns, 1000)):
      chunkRes = self.fcClient.getReplicas(chunk)
      if not chunkRes['OK']:
        self.logError('Failed to get replicaInformation: ', chunkRes['Message'])
        return chunkRes
      self.log.info('Chunk %s/%s done' % (index, len(lfns) // 1000))
      res['Successful'].update(chunkRes['Value']['Successful'])
      res['Failed'].update(chunkRes['Value']['Failed'])

    result = {}
    result['Successful'] = {}
    result['Failed'] = {}
    setOfSEs = set(storageElements)

    for lfn, msg in res['Failed'].items():
      if msg == 'No such file or directory':
        result['Successful'][lfn] = False
      else:
        result['Failed'][lfn] = msg

    # check if all replicas are registered in FC
    filesFoundInFC = res['Successful']
    for lfn, replicas in filesFoundInFC.items():
      result['Successful'][lfn] = setOfSEs.issubset(set(replicas))

    return S_OK(result)

  def existsOnSE(self, storageElements, lfns, transInfoDict):
    """checks if the given files exist physically on a list of storage elements."""

    result = {}
    result['Failed'] = {}
    result['Successful'] = {}
    authorDN = transInfoDict['AuthorDN']
    authorGroup = transInfoDict['AuthorGroup']

    if not lfns:
      return S_OK(result)

    voName = lfns[0].split('/')[1]
    with UserProxy(proxyUserDN=authorDN, proxyUserGroup=authorGroup) as proxyResult:
      if not proxyResult['OK']:
        return S_ERROR('Failed to get a proxy: %s' % proxyResult['Message'])
      for se in storageElements:
        for index, lfnChunk in enumerate(breakListIntoChunks(lfns, 200)):
          self.log.notice('Checking LFNs at', '%s: %s/%s' % (se, index, len(lfns) / len(lfnChunk)))
          res = StorageElement(se, vo=voName).exists(lfnChunk)
          if not res['OK']:
            return res
          for lfn, status in res['Value']['Successful'].items():
            if lfn not in result['Successful']:
              result['Successful'][lfn] = status
            if not status:
              result['Successful'][lfn] = False
          result['Failed'].setdefault(se, {}).update(res['Value']['Failed'])

    return S_OK(result)

  def exists(self, storageElements, lfns, transInfoDict):
    """checks if files exists on both file catalog and storage elements."""

    fcRes = self.existsInFC(storageElements, lfns)
    if not fcRes['OK']:
      self.logError('Failure to determine if files exists in File Catalog: ', fcRes['Message'])
      return fcRes

    if fcRes['Value']['Failed']:
      self.logError("Failed FileCatalog Response ", "%s" % fcRes['Value']['Failed'])

    # check if files found in file catalog also exist on SE
    checkLFNsOnStorage = [lfn for lfn in fcRes['Value']['Successful'] if fcRes['Value']['Successful'][lfn]]

    # no files were found in FC, return the result instead of verifying them on SE
    if not checkLFNsOnStorage:
      return fcRes

    seRes = self.existsOnSE(storageElements, checkLFNsOnStorage, transInfoDict)
    if not seRes['OK']:
      self.logError('Failure to determine if files exist on SE ', "%s" % seRes['Message'])
      return seRes

    for se in storageElements:
      if seRes['Value']['Failed'][se]:
        self.logError('Failed to determine if files exist on SE ', "%s %s" % (se, seRes['Value']['Failed'][se]))
        return S_ERROR()

    fcResult = fcRes['Value']['Successful']
    seResult = seRes['Value']['Successful']
    for lfn in fcResult:
      if fcResult[lfn] and not seResult[lfn]:
        fcRes['Value']['Successful'][lfn] = False

    return fcRes

  def processTransformation(self, transID, sourceSE, targetSEs, transType, transInfoDict):
    """process transformation for a given transformation ID."""

    actions = {}
    actions[SET_PROCESSED] = []
    actions[RETRY] = []
    actions[SET_DELETED] = []

    for status in self.transformationFileStatuses:
      res = self.tClient.getTransformationFiles(condDict={'TransformationID': transID, 'Status': status})
      if not res['OK']:
        errStr = 'Failure to get Transformation Files, Status: %s Transformation ID: %s Message: %s' % (status,
                                                                                                        transID,
                                                                                                        res['Message'])
        self.logError(errStr)
        continue

      transFiles = res['Value']
      if not transFiles:
        self.log.notice("No Transformation Files found with status %s for Transformation ID %d" % (status, transID))
        continue

      self.log.notice("Processing Transformation Files with status %s for TransformationID %d " % (status, transID))

      if status == 'Assigned':
        transFiles = list(filter(self.selectFailedRequests, transFiles))

      lfns = [transFile['LFN'] for transFile in transFiles]

      if not lfns:
        continue

      res = self.exists(sourceSE, lfns, transInfoDict)
      if not res['OK']:
        continue

      resultSourceSe = res['Value']['Successful']

      res = self.exists(targetSEs, lfns, transInfoDict)
      if not res['OK']:
        continue
      resultTargetSEs = res['Value']['Successful']

      for transFile in transFiles:
        lfn = transFile['LFN']
        transFile['AvailableOnSource'] = resultSourceSe[lfn]
        transFile['AvailableOnTarget'] = resultTargetSEs[lfn]

      checkFilesFuncName = "check_%s_files" % status.lower()
      checkFiles = getattr(self, checkFilesFuncName)
      checkFiles(actions, transFiles, transType)

    self.applyActions(transID, actions)
    self.sendNotification(transID, transType, sourceSE, targetSEs)

    return S_OK()



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Set of functions used to resolve the applications' dependencies, looking into the CS.

Works recursively

:since: Apr 26, 2010
:author: Stephane Poss
"""

from __future__ import absolute_import
__RCSID__ = "$Id$"

from DIRAC import gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations


def resolveDeps(sysconfig, appli, appversion):
  """Resolve the dependencies.

  :param str sysconfig: system configuration
  :param str appli: application name
  :param str appversion: application version
  :return: list of dictionaries
  """
  log = gLogger.getSubLogger("resolveDeps")
  ops = Operations()
  deps = ops.getSections('/AvailableTarBalls/%s/%s/%s/Dependencies' % (sysconfig, appli,
                                                                       appversion), '')
  depsarray = []
  if deps['OK']:
    for dep in deps['Value']:
      vers = ops.getValue('/AvailableTarBalls/%s/%s/%s/Dependencies/%s/version' % (sysconfig, appli,
                                                                                   appversion, dep), '')
      depvers = ''
      if vers:
        depvers = vers
      else:
        log.error("Retrieving dependency version for %s failed, skipping to next !" % (dep))
        continue
      log.verbose("Found dependency %s %s" % (dep, depvers))
      depdict = {}
      depdict["app"] = dep
      depdict["version"] = depvers
      depsarray.append(depdict)
      # resolve recursive dependencies
      depsofdeps = resolveDeps(sysconfig, dep, depvers)
      depsarray.extend(depsofdeps)
  else:
    log.verbose("Could not find any dependency for %s %s, ignoring" % (appli, appversion))
  return depsarray



#!/usr/local/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test SLIC module."""

from __future__ import absolute_import
import unittest
from mock import patch, MagicMock as Mock

from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Interfaces.API.NewInterface.Applications import SLIC
from Tests.Utilities.GeneralUtils import assertEqualsImproved, assertDiracFailsWith, \
    assertDiracSucceeds, assertInImproved

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Interfaces.API.NewInterface.Applications.SLIC'

# pylint: disable=protected-access


class SLICTestCase(unittest.TestCase):
  """Base class for the SLIC test cases."""

  def setUp(self):
    """set up the objects."""
    self.slic = SLIC({})
    self.slic.version = 941

  def test_setters(self):
    self.slic.setRandomSeed('someInvalidSeed')
    self.slic.setDetectorModel('myDetModelTestme')
    self.slic.setStartFrom('soirmgf')
    assertInImproved('_checkArgs', self.slic._errorDict, self)
    (rand_found, start_found) = (False, False)
    for err in self.slic._errorDict['_checkArgs']:
      if 'randomSeed' in err:
        rand_found = True
      elif 'startfrom' in err:
        start_found = True
    assertEqualsImproved((True, True, 'myDetModelTestme'), (rand_found, start_found,
                                                                self.slic.detectorModel), self)

  def test_setdetectormodel(self):
    self.slic.setDetectorModel('lfn:/my/det.model')
    assertInImproved('lfn:/my/det.model', self.slic.inputSB, self)
    assertEqualsImproved(self.slic.detectorModel, 'det.model', self)
    self.slic = SLIC({})
    with patch('os.path.exists', new=Mock(return_value=True)):
      self.slic.setDetectorModel('/my/local/dir/detectorv212.stdhep.zip')
      assertEqualsImproved((self.slic.detectorModel, self.slic.inputSB),
                            ('detectorv212.stdhep',
                              ['/my/local/dir/detectorv212.stdhep.zip']), self)
    self.slic = SLIC({})
    with patch('os.path.exists', new=Mock(return_value=False)):
      self.slic.setDetectorModel('/my/local/dir/detectorv212.stdhep.zip')
      assertEqualsImproved((self.slic.detectorModel, self.slic.inputSB),
                            ('detectorv212.stdhep', []), self)

  def test_userjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.slic._userjobmodules(module_mock), self)

  def test_prodjobmodules(self):
    module_mock = Mock()
    assertDiracSucceeds(self.slic._prodjobmodules(module_mock), self)

  def test_userjobmodules_fails(self):
    with patch('%s._setUserJobFinalization' % MODULE_NAME, new=Mock(return_value=S_OK('something'))),\
        patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_test_err'))):
      assertDiracFailsWith(self.slic._userjobmodules(None), 'userjobmodules failed', self)

  def test_prodjobmodules_fails(self):
    with patch('%s._setApplicationModuleAndParameters' % MODULE_NAME, new=Mock(return_value=S_OK('something'))), \
        patch('%s._setOutputComputeDataList' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_other_test_err'))):
      assertDiracFailsWith(self.slic._prodjobmodules(None), 'prodjobmodules failed', self)

  def test_checkconsistency(self):
    self.slic._jobtype = 'User'
    self.slic.startFrom = True
    with patch('os.path.exists', new=Mock(return_value=True)):
      assertDiracSucceeds(self.slic._checkConsistency(), self)

  def test_checkconsistency_othercase(self):
    self.slic.detectorModel = 'someModel'
    self.slic.steeringFile = 'LFN:/my/consistent/steeringfile.st'
    self.slic._jobtype = 'notUser'
    self.slic.startFrom = False
    with patch('os.path.exists', new=Mock(return_value=False)):
      assertDiracSucceeds(self.slic._checkConsistency(), self)

  def test_checkconsistency_download_succeeds(self):
    import inspect
    self.slic.steeringFile = '/my/remote/path.st'
    self.slic._jobtype = 'notUser'
    self.slic.startFrom = False
    with patch('os.path.exists', new=Mock(return_value=False)), \
         patch.object(inspect.getmodule(SLIC), 'Exists', new=Mock(return_value=S_OK('ok'))):
      assertDiracSucceeds(self.slic._checkConsistency(), self)

  def test_checkconsistency_noversion(self):
    self.slic.version = None
    assertDiracFailsWith(self.slic._checkConsistency(), 'no version found', self)

  def test_checkconsistency_nosteeringfile(self):
    import inspect
    self.slic.steeringFile = 'mysteer.file'
    with patch('os.path.exists', new=Mock(return_value=False)), \
         patch.object(inspect.getmodule(SLIC), 'Exists', new=Mock(return_value=S_ERROR('failed downloading, testerr'))):
      assertDiracFailsWith(self.slic._checkConsistency(), 'failed downloading, testerr', self)

  def test_resolvestepparams(self):
    step_mock = Mock()
    inputstep_mock = Mock()
    self.slic._linkedidx = 2
    self.slic._jobsteps = [None, None, inputstep_mock]
    assertDiracSucceeds(self.slic._resolveLinkedStepParameters(step_mock), self)

    assertDiracSucceeds(self.slic._checkWorkflowConsistency(), self)

  def test_resolvestepparams_noinputstep(self):
    self.slic._linkedidx = '139'
    self.slic._jobsteps = []
    assertDiracSucceeds(self.slic._resolveLinkedStepParameters(''), self)



#!/usr/local/env python
#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test ReleaseHelper module."""

from __future__ import absolute_import
import unittest
from mock import patch, MagicMock as Mock

from ILCDIRAC.ILCTransformationSystem.Utilities.ReleaseHelper import killRPath, getFiles, \
    copyLibraries, resolveLinks, getLibraryPath, copyFolder, getPythonStuff, removeSystemLibraries, \
    getDependentLibraries, getGeant4DataFolders, getRootStuff, insertCSSection
from Tests.Utilities.GeneralUtils import assertEqualsImproved, assertDiracFailsWith, \
    assertDiracSucceeds, assertMockCalls, assertListContentEquals

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.ILCTransformationSystem.Utilities.ReleaseHelper'

# pylint: disable=too-many-public-methods,no-self-use


class ReleaseHelperTestCase(unittest.TestCase):
  """Base class for the ReleaseHelper test cases."""

  def setUp(self):
    """set up the objects."""
    pass

#  def tearDown( self ):
#    pass

  def test_killrpath(self):
    walk_list = [('/testRoot', 'nodirs', ['file1.txt', 'deletethislib.so', '', 'ignorefile.s o']),
                  ('/otherRoot/mydir/', '', ['lib1.so', 'subfolder/otherlib.so', 'library.dll'])]
    with patch('%s.os.walk' % MODULE_NAME, new=Mock(return_value=walk_list)) as walk_mock, \
         patch('%s.getStatusOutput' % MODULE_NAME, new=Mock()) as cmd_mock:
      killRPath('myTest_folder')
      walk_mock.assert_called_once_with('myTest_folder', followlinks=True)
      expected_files = ['/testRoot/deletethislib.so', '/otherRoot/mydir/lib1.so',
                         '/otherRoot/mydir/subfolder/otherlib.so']
      assertMockCalls(cmd_mock, ['chrpath -d ' + x + ' ' for x in expected_files], self)

  def test_killrpath_donothing(self):
    with patch('%s.os.walk' % MODULE_NAME, new=Mock(return_value=[])) as walk_mock, \
        patch('%s.getStatusOutput' % MODULE_NAME, new=Mock()) as cmd_mock:
      killRPath('myTest_folder')
      walk_mock.assert_called_once_with('myTest_folder', followlinks=True)
      self.assertFalse(cmd_mock.called)

  def test_getfiles(self):
    walk_list = [('/testRoot', 'nodirs', ['file1.txt', 'deletethislib.so', '', 'ignorefile.stdhep']),
                  ('/otherRoot/mydir/', '', ['particle.stdhep', 'subfolder/otherlib.stdhep', 'library.dll'])]
    with patch('%s.os.walk' % MODULE_NAME, new=Mock(return_value=walk_list)) as walk_mock:
      res = getFiles('myTest_folder', '.stdhep')
      walk_mock.assert_called_once_with('myTest_folder', followlinks=True)
      expected_files = ['/testRoot/ignorefile.stdhep', '/otherRoot/mydir/particle.stdhep',
                         '/otherRoot/mydir/subfolder/otherlib.stdhep']
      assertListContentEquals(res, expected_files, self)

  def test_getfiles_donothing(self):
    with patch('%s.os.walk' % MODULE_NAME, new=Mock(return_value=[])) as walk_mock:
      self.assertFalse(getFiles('myTest_folder', ''))
      walk_mock.assert_called_once_with('myTest_folder', followlinks=True)

  def test_copylibraries(self):
    with patch('%s.getStatusOutput' % MODULE_NAME, new=Mock(return_value=(0, '', ''))) as cmd_mock:
      copyLibraries(['myFile.txt', 'some_dir/otherfile.so', 'copythis.log'], 'testFolderTarget')
      cmd_mock.assert_called_once_with(
          "rsync --exclude '.svn' -avzL  myFile.txt some_dir/otherfile.so copythis.log testFolderTarget ")

  def test_copylibraries_copy_error(self):
    with patch('%s.getStatusOutput' % MODULE_NAME, new=Mock(return_value=(1, '', ''))), \
        self.assertRaises(RuntimeError):
      copyLibraries(['myFile.txt', 'some_dir/otherfile.so', 'copythis.log'], 'testFolderTarget')

  def test_resolvelinks(self):
    walk_list = [('/my/test/rootdir', 'ignored',
                    ['testfile_1.txt', 'mylib_a.so', 'subfolder/otherlib_b.so.1', 'mylib_a.so.1']),
                  ('/other/libfolder/', 'other_ignore', ['library.so', 'requirement.so', 'otherlib_b.so',
                                                           'requirement.so.2']), ('', '', [])]
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/test/dir/cwd')) as getcwd_mock, \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock()) as chdir_mock, \
         patch('%s.os.walk' % MODULE_NAME, new=Mock(return_value=walk_list)) as walk_mock, \
         patch('%s.os.remove' % MODULE_NAME, new=Mock()) as remove_mock, \
         patch('%s.os.symlink' % MODULE_NAME, new=Mock()) as slink_mock:
      resolveLinks('my_test_dir')
      getcwd_mock.assert_called_once_with()
      assertMockCalls(chdir_mock, ['my_test_dir', '/test/dir/cwd'], self)
      assertMockCalls(remove_mock, ['mylib_a.so', 'otherlib_b.so', 'requirement.so'], self)
      assertMockCalls(slink_mock, [('mylib_a.so.1', 'mylib_a.so'), ('otherlib_b.so.1', 'otherlib_b.so'),
                                     ('requirement.so.2', 'requirement.so')], self)
      walk_mock.assert_called_once_with('my_test_dir', followlinks=True)

  def test_resolvelinks_none_match(self):
    walk_list = [('/my/test/rootdir', 'ignored', ['testfile_1.txt', 'mylib_a.so', 'subfolder/otherlib_b.so.1']),
                  ('/other/libfolder/', 'other_ignore', ['library.so', 'requirement.so']), ('', '', [])]
    with patch('%s.os.getcwd' % MODULE_NAME, new=Mock(return_value='/test/dir/cwd')) as getcwd_mock, \
         patch('%s.os.chdir' % MODULE_NAME, new=Mock()) as chdir_mock, \
         patch('%s.os.walk' % MODULE_NAME, new=Mock(return_value=walk_list)), \
         patch('%s.os.remove' % MODULE_NAME, new=Mock()) as remove_mock, \
         patch('%s.os.symlink' % MODULE_NAME, new=Mock()) as slink_mock:
      resolveLinks('my_test_dir')
      getcwd_mock.assert_called_once_with()
      assertMockCalls(chdir_mock, ['my_test_dir', '/test/dir/cwd'], self)
      self.assertFalse(remove_mock.called)
      self.assertFalse(slink_mock.called)

  def test_getlibrarypath(self):
    assertEqualsImproved(getLibraryPath('/some/folder/on/the/fs'), '/some/folder/on/the/fs/lib', self)

  def test_getlibrarypath_withslash(self):
    assertEqualsImproved(getLibraryPath('/some/folder/on/the/fs/'), '/some/folder/on/the/fs/lib', self)

  def test_getlibrarypath_nopath(self):
    assertEqualsImproved(getLibraryPath(''), 'lib', self)

  def test_copyfolder(self):
    with patch('%s.os.makedirs' % MODULE_NAME, new=Mock()) as makedirs_mock, \
        patch('%s.getStatusOutput' % MODULE_NAME, new=Mock(return_value=(0, '', ''))) as cmd_mock:
      copyFolder('test_base', 'test_destination/some/folders')
      makedirs_mock.assert_called_once_with('test_destination/some/folders')
      cmd_mock.assert_called_once_with("rsync --exclude '.svn' -avzL  test_base test_destination/some/folders ")

  def test_copyfolder_ignore_oserr(self):
    with patch('%s.os.makedirs' % MODULE_NAME, new=Mock(side_effect=OSError('test_err'))) as makedirs_mock, \
        patch('%s.getStatusOutput' % MODULE_NAME, new=Mock(return_value=(0, '', ''))) as cmd_mock:
      copyFolder('test_base', 'test_destination/some/folders')
      makedirs_mock.assert_called_once_with('test_destination/some/folders')
      cmd_mock.assert_called_once_with("rsync --exclude '.svn' -avzL  test_base test_destination/some/folders ")

  def test_copyfolder_copy_fails(self):
    with patch('%s.os.makedirs' % MODULE_NAME, new=Mock()) as makedirs_mock, \
        patch('%s.getStatusOutput' % MODULE_NAME, new=Mock(return_value=(1, '', ''))) as cmd_mock, \
        self.assertRaises(RuntimeError):
      copyFolder('test_base', 'test_destination/some/folders')
      makedirs_mock.assert_called_once_with('test_destination/some/folders')
      cmd_mock.assert_called_once_with("rsync --exclude '.svn' -avzL  test_base test_destination/some/folders ")

  def test_getpythonstuff(self):
    with patch('%s.copyFolder' % MODULE_NAME, new=Mock()) as copy_mock:
      getPythonStuff('source', 'dst')
      copy_mock.assert_called_once_with('source', 'dst')

  def test_removesyslibs(self):
    walk_list = [('/my/usr/lib/', 'ignore',
                    ['', 'libc.so.2', 'myfile.txt', 'libc-2.5.so', 'libm.so.1', 'libpthread.so.123',
                      'ignore_me.log', 'libdl.so', 'ignore_me.so.1', 'libstdc++.so.1', 'ignore_me.so',
                      'libgcc_s.so.1']),
                  ('/lib/dir/subdir/', 'ignoretoo',
                    ['some_file.so', 'other_lib.so.1', 'libc.so', 'libc-2.5.so.1', 'libm.so',
                      'libpthread.so', 'libdl.so.124', 'libstdc++.so', 'libgcc_s.so.1.18'])]
    remove_vals = [None, OSError('test_remove_failure')] + [None] * 12
    with patch('%s.os.walk' % MODULE_NAME, new=Mock(return_value=walk_list)) as walk_mock, \
         patch('%s.os.remove' % MODULE_NAME, new=Mock(side_effect=remove_vals)) as remove_mock:
      removeSystemLibraries('my_folder_tbr')
      walk_mock.assert_called_once_with('my_folder_tbr')
      expected_remove_files_1 = ['libc.so.2', 'libc-2.5.so', 'libm.so.1', 'libpthread.so.123',
                                  'libdl.so', 'libstdc++.so.1', 'libgcc_s.so.1']
      expected_remove_files_2 = ['libc.so', 'libc-2.5.so.1', 'libm.so', 'libpthread.so', 'libdl.so.124',
                                  'libstdc++.so', 'libgcc_s.so.1.18']
      assertMockCalls(remove_mock, ['/my/usr/lib/' + x for x in expected_remove_files_1]
                       + ['/lib/dir/subdir/' + x for x in expected_remove_files_2], self)

  def test_getdependentlibs(self):
    outputlines = '\tsome_string=> some_replacement (0x\n\ttest_string_deleteme=> valid_pattern         asdgf           (0x\n\n some_other_string=> invalid_replacement(0x\n'
    with patch('%s.getStatusOutput' % MODULE_NAME, new=Mock(return_value=(0, outputlines, ''))) as cmd_mock:
      res = getDependentLibraries('MyTest_library')
      cmd_mock.assert_called_once_with('ldd MyTest_library')
      assertEqualsImproved(res, {'some_replacement', 'valid_pattern         asdgf          '}, self)

  def test_getdependentlibs_fails(self):
    outputlines = '\tsome_string=> some_replacement (0x\n\t not found 98jefeiufmn=> aufm (0x\n\ttest_string_deleteme=> valid_pattern         asdgf           (0x\n\n'
    with patch('%s.getStatusOutput' % MODULE_NAME, new=Mock(return_value=(0, outputlines, ''))), \
         self.assertRaises(RuntimeError):
      getDependentLibraries('MyTest_library')

  def test_getgeant4datafolders(self):
    import os
    with patch('%s.copyFolder' % MODULE_NAME, new=Mock()) as copy_mock, \
         patch.dict(os.environ, {'MY_TEST_VAR': '/my/test/path/dir'}):
      getGeant4DataFolders('MY_TEST_VAR', '/test_target_folder/fold/')
      copy_mock.assert_called_once_with('/my/test/path/dir', '/test_target_folder/fold/')

  def test_getrootstuff(self):
    file_list = ['/my_test/aim/lib/some_lib.so', '/my_test/aim/lib/my_important_lib.so.1',
                  '/my_test/aim/lib/last_lib.so.123']
    dependency_list = [{'first_dependency.so', 'base_lib.so.1'}, {}, {'last_dependency.so.39'}]
    with patch('%s.getStatusOutput' % MODULE_NAME, new=Mock(return_value=(0, '', ''))) as cmd_mock, \
         patch('%s.getFiles' % MODULE_NAME, new=Mock(return_value=file_list)) as getfile_mock, \
         patch('%s.getDependentLibraries' % MODULE_NAME, new=Mock(side_effect=dependency_list)) as getlibs_mock, \
         patch('%s.copyLibraries' % MODULE_NAME, new=Mock()) as copy_mock:
      getRootStuff('test_rootsys', '/my_test/aim')
      cmd_mock.assert_called_once_with(
          "rsync --exclude '.svn' -av test_rootsys/lib test_rootsys/etc test_rootsys/bin test_rootsys/cint  /my_test/aim")
      getfile_mock.assert_called_once_with('/my_test/aim/lib', '.so')
      assertMockCalls(getlibs_mock, [
          '/my_test/aim/lib/some_lib.so', '/my_test/aim/lib/my_important_lib.so.1',
          '/my_test/aim/lib/last_lib.so.123'], self)
      copy_mock.assert_called_once_with({'first_dependency.so', 'base_lib.so.1', 'last_dependency.so.39'},
                                         '/my_test/aim/lib')

  def test_getrootstuff_fails(self):
    with patch('%s.getStatusOutput' % MODULE_NAME, new=Mock(return_value=(1, '', ''))), \
        self.assertRaises(RuntimeError):
      getRootStuff('rootsys', 'target')

  def test_getcssection(self):
    from DIRAC import S_OK
    cs_mock = Mock()
    cs_mock.setOption.return_value = S_OK()
    assertDiracSucceeds(
        insertCSSection(cs_mock, '/TestConfig/full/path', {
            'newsubsection/Parameter1': '194851', 'newsubsection':
            {'newsubsubsection/ComplexParameter': True}}), self)
    assertMockCalls(cs_mock.setOption, [
        ('/TestConfig/full/path/newsubsection/Parameter1', '194851'),
        ('/TestConfig/full/path/newsubsection/newsubsubsection/ComplexParameter', True)], self)
    assertMockCalls(cs_mock.createSection, ['/TestConfig/full/path', '/TestConfig/full/path',
                                              '/TestConfig/full/path/newsubsection'], self)

  def test_getcssection_fails(self):
    from DIRAC import S_OK, S_ERROR
    cs_mock = Mock()
    cs_mock.setOption.side_effect = [S_OK(), S_ERROR('set_option_failed!_testme')]
    assertDiracFailsWith(
        insertCSSection(cs_mock, '/TestConfig/full/path',
                        {'newsubsection/Parameter1': '194851',
                         'newsubsection/Parameter2_invalid_value': 'i cause an error'}),
        'set_option_failed!_testme', self)
